import sys
sys.path.append('.')
from colabfit.tools.dataset_old import Dataset, load_data
from colabfit.tools.property_settings import PropertySettings
from colabfit.tools.transformations import ExtractCauchyStress

# Define transformations to use when parsing data
T = {'stress': ExtractCauchyStress()}

# Create a skeleton Dataset
ds1 = Dataset('InP_JPCA2020')

ds1.authors = ['M. A. Cusentino', 'M. A. Wood', 'A. P. Thompson']
ds1.links = [
    'https://pubs.acs.org/doi/10.1021/acs.jpca.0c02450',
    'https://github.com/FitSNAP/FitSNAP/tree/master/examples/InP_JPCA2020'
]
ds1.description  = \
    'This data set was used to generate a multi-element linear '\
    'SNAP potential for InP as published in Cusentino, M.A. et. al, J. Chem. '\
    'Phys. (2020). Intended to produce an interatomic potential for '\
    'indium phosphide capable of capturing high-energy defects that '\
    'result from radiation damage cascades.'

# Add configurations, with their linked data if it exists
ds1.configurations = load_data(
    file_path='/home/josh/colabfit/data/FitSNAP/InP_JPCA2020.extxyz',
    file_format='xyz',
    name_field='name',
    elements=['In', 'P'],
    default_name=ds1.name
)

# Rename a property name to a known name
ds1.rename_property('virial', 'stress')

ds1.property_map = {
    # ColabFit name: {'field': ASE field name, 'units': str}
    'energy': {'field': 'energy', 'units': 'eV'},
    'forces': {'field': 'forces', 'units': 'eV/Ang'},
    'stress': {'field': 'stress', 'units': 'kilobar'},
}

# Extract the properties from the configurations and loads them to Dataset.data
ds1.parse_data(
    transformations=T,
    convert_units=False
)

# Build configuration sets by setting the cs_regexes dictionary
ds1.cs_regexes = {
    'default':
        'Curated configurations for producing an interatomic '\
        'potential for indium phosphide capable of capturing high-energy '\
        'defects that result from radiation damage cascades',
    '^Bulk':
        'Ground state configuration for bulk zinc blende',
    '^EOS':
        'Bulk zinc blende with uniform expansion and compression',
    '^Shear':
        'Bulk zincblende with random cell shape modifications',
    '^Strain':
        'Uniaxially strained bulk zinc blende',
    '^a(In|P)':
        'Antisite defects in InP',
    '^aa':
        'Diantisite defects',
    '^i(In|P)':
        'Interstitial defects in InP',
    '^vP':
        'Vacancy defects in InP',
    '^vv':
        'Divacancy defects in InP',
    '^s_a(In|P|a)':
        'No description',
    '^s_i(In|P)':
        'No description',
    '^s_v(In|P)':
        'No description',
    '^s_vv':
        'No description',
}

# Apply configuration labels by setting the cs_regexes dictionary
ds1.co_label_regexes = {
    'Bulk|EOS|Shear|Strain':
        'zincblende',
    'EOS':
        'eos',
    'Shear|Strain':
        'strain',
    '^a(In|P)':
        'antisite',
    '^aa':
        'diantisite',
    '^i(In|P)':
        'interstitial',
    '^v(In|P|v)':
        'vacancy',
}

ds1.ps_regexes = {
    '.*':
        PropertySettings(
            method='VASP',
            description='energies/forces/stresses',
            files=None,
            labels=['PBE', 'GGA'],
        )
}

# Resync to make sure metadata is updated everywhere necessary
ds1.resync()

print(ds1)

ds1.to_markdown(
    'tests/files',
    'dummy_written1.md',
    'dummy_written1.extxyz',
    'xyz'
)

# Create a skeleton Dataset
ds2 = Dataset('MoNbTaVW_PRB2021')

ds2.authors      = ['J. Byggmästar', 'K. Nordlund', 'F. Djurabekova']
ds2.links = [
    'https://journals.aps.org/prb/abstract/10.1103/PhysRevB.104.104101',
    'https://doi.org/10.23729/1b845398-5291-4447-b417-1345acdd2eae'
]
ds2.description  = \
    'This dataset was originally designed to fit a GAP model '\
    'for the Mo-Nb-Ta-V-W quinary system that was used to study segregation '\
    'and defects in the body-centered-cubic refractory high-entropy alloy '\
    'MoNbTaVW.'

# Add configurations, with their linked data if it exists
ds2.configurations = load_data(
    file_path='/home/josh/colabfit/data/acclab_helsinki/MoNbTaVW_cleaned.xyz',
    file_format='xyz',
    name_field='name',
    elements=['Mo', 'Nb', 'Ta', 'V', 'W'],
    default_name=ds2.name
)

# Rename a property name to a known name
ds2.rename_property('virial', 'stress')

ds2.property_map = {
    # ColabFit name: {'field': ASE field name, 'units': str}
    'energy': {'field': 'energy', 'units': 'eV'},
    'forces': {'field': 'force',  'units': 'eV/Ang'},
    'stress': {'field': 'stress', 'units': 'kilobar'},
}

# Extract the properties from the configurations
ds2.parse_data(
    transformations=T,
    convert_units=False
)

# Build configuration sets by setting the cs_regexes dictionary
ds2.cs_regexes = {
    'default':
        'A variety of Mo-Nb-Ta-V-W structures',
    'bcc_distorted':
        'BCC configurations with random strains up to +/- 30% to '\
        'help train the far-from-equilibrium elastic response',
    'binary_alloys':
        'Binary BCC alloys sampling 10 different concentrations '\
        'from A_0.05B_0.95 to A_0.95B_0.05 and 3 different lattice constants '\
        'for every composition. Atoms are randomly ordered and shifted '\
        'slightly from their lattice positions.',
    '^composition':
        'Ternary, quaternary, and quinary BCC alloys. 3 linearly '\
        'spaced compositions were sampled, each with 3 different lattice '\
        'constants. Atoms are randomly ordered and shifted slightly from '\
        'their lattice positions.',
    'di-sia':
        'Configurations with two self-interstitial atoms',
    'di-vacancy':
        'Configurations with two vacancies',
    'dimer':
        'Dimers for fitting repulsive terms',
    'gamma_surface':
        'Configurations representing the full gamma surface',
    'hea_ints':
        '1-5 interstitial atoms randomly inserted into HEA lattices and '\
        'relaxed with a partially-trained tabGAP model',
    'hea_short_range':
        'Randomly placed unrelaxed interstitial atom in HEAs '\
        'to fit repulsion inside crystals, making sure that the closest '\
        'interatomic distance is not too short for DFT to be unreliable '\
        '(> 1.35 Ang)',
    'hea_small':
        'Bulk equiatomic quinary HEAs. Atoms are randomly ordered and shifted '\
        'slightly from their lattice positions. The lattice constant is '\
        'randomised in the range 3-3.4 Angstrom',
    'hea_surface':
        'Disordered HEA surfaces, including some of the '\
        'damaged/molten surface configurations from an existing pure W '\
        'dataset that were turned into HEAs',
    'hea_vacancies':
        '1-5 vacancies randomly inserted into HEA lattices, then '\
        'relaxed with a partially-trained tabGAP model',
    'isolated_atom':
        'Isolated atoms',
    '^liquid__':
        'Liquid configurations',
    'liquid_composition':
        'Liquid equiatomic binary, ternary, quaternary, and quinary alloys at '\
        'different densities',
    'liquid_hea':
        'Liquid HEA configurations',
    'mcmd':
        'Equiatomic quinary alloys generated via active learning by running '\
        'MCMD with a partially-trained tabGAP model.',
    'ordered_alloys':
        'Ordered binary, ternary, and quaternary alloys (always as a BCC '\
        'lattice, but with different crystal symmetries of the elemental '\
        'sublattices)',
    'phonon':
        'MD snapshots taken at 1000K for three different volumes',
    '^short_range':
        'BCC crystals with random interstitial atom defects to capture '\
        'short-range many-body dynamics',
    '^sia':
        'Configurations with a self-interstitial defect',
    'surf_liquid':
        'Damaged and molten surfaces',
    'tri-vacancy':
        'Configurations with three vacancies',
    '^vacancy':
        'Configurations with one vacancy',
}

# Apply configuration labels by setting the cs_regexes dictionary
ds2.co_label_regexes = {
    'alloys':
        'bcc',
    'bcc':
        'bcc strain',
    'di-vacancy':
        ['vacancy', 'divacancy'],
    'dimer':
        ['dimer', 'warning', 'large_forces', 'repulsive'],
    'gamma_surface':
        'gamma_surface',
    'hea':
        'hea',
    'hea_ints':
        'interstitial',
    'isolated_atom':
        'isolated_atom',
    'liquid':
        'liquid',
    'mcmd':
        'aimd',
    'phonon':
        'aimd',
    'short_range':
        ['bcc', 'interstitial', 'repulsive', 'large_forces', 'warning'],
    'sia':
        'interstitial',
    'surface':
        'surface',
    'tri-vacancy':
        ['vacancy', 'divacancy', 'trivacancy'],
    'vacancies':
        'vacancy',
    'vacancy':
        'vacancy',
}

ds2.ps_regexes = {
    '.*':
        PropertySettings(
            method='VASP',
            description='energies/forces/stresses',
            files=None,
            labels=['PBE', 'GGA'],
        )
}

# Resync to make sure metadata is updated everywhere necessary
ds2.resync()

ds2.to_markdown(
    'tests/files/',
    'dummy_written2.md',
    'dummy_written2.extxyz',
    'xyz'
)

ds1 = Dataset.from_markdown('tests/files/dummy_written1.md', transformations=T)
ds2 = Dataset.from_markdown('tests/files/dummy_written2.md', transformations=T)

# print(ds1)
# print(ds2)

nested_ds = Dataset('nested-ds')

nested_ds.attach_dataset(ds1)
nested_ds.attach_dataset(ds2)

nested_ds.resync()

print(nested_ds)

nested_ds.to_markdown(
    'tests/files/',
    'dummy_nested.md',
    'dummy_nested.extxyz',
    'xyz')


import os
import pytest
import tempfile
import numpy as np

from ase import Atoms

from colabfit import (
    ATOMS_NAME_FIELD, ATOMS_LABELS_FIELD
)
from colabfit.tools.hdf5_backend import ConcatenationException, HDF5Backend
from colabfit.tools.configuration import Configuration
from colabfit.tools.property_settings import PropertySettings

def build_n(n):
    images              = []
    energies            = []
    stress              = []
    names               = []
    nd_same_shape       = []
    nd_diff_shape       = []
    forces              = []
    nd_same_shape_arr   = []
    nd_diff_shape_arr   = []

    for i in range(1, n+1):
        atoms = Atoms(f'H{i}', positions=np.random.random((i, 3)))

        atoms.info['energy'] = np.random.random()
        atoms.info['stress'] = np.random.random(6)
        atoms.info['name'] = f'configuration_{i}'
        atoms.info['nd-same-shape'] = np.random.random((2, 3, 5))
        atoms.info['nd-diff-shapes'] = np.random.random((
            i+np.random.randint(1, 4),
            i+1+np.random.randint(1, 4),
            i+2+np.random.randint(1, 4),
        ))

        energies.append(atoms.info['energy'])
        stress.append(atoms.info['stress'])
        names.append(atoms.info['name'])
        nd_same_shape.append(atoms.info['nd-same-shape'])
        nd_diff_shape.append(atoms.info['nd-diff-shapes'])

        atoms.arrays['forces'] = np.random.random((i, 3))
        atoms.arrays['nd-same-shape-arr'] = np.random.random((i, 2, 3))
        atoms.arrays['nd-diff-shapes-arr'] = np.random.random((
            i,
            i+np.random.randint(1, 4),
            i+1+np.random.randint(1, 4),
        ))

        forces.append(atoms.arrays['forces'])
        nd_same_shape_arr.append(atoms.arrays['nd-same-shape-arr'])
        nd_diff_shape_arr.append(atoms.arrays['nd-diff-shapes-arr'])

        images.append(Configuration.from_ase(atoms))

    return (
        images,
        energies, stress, names, nd_same_shape, nd_diff_shape,
        forces, nd_same_shape_arr, nd_diff_shape_arr,
    )


class TestAddingConfigurations:

    def test_add_then_update_nochange_config(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = HDF5Backend(tmpfile, mode='w')

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            database.insert_data(images)
            database.insert_data(images)

            for co_data in database.get_data('configurations/names').values():
                assert len(co_data) == 0

            for co_data in database.get_data('configurations/labels').values():
                assert len(co_data) == 0

            for co_data in database.get_data('configurations/last_modified').values():
                assert co_data.shape == ()

            database.concatenate_group('configurations/atomic_numbers')
            database.concatenate_group('configurations/positions')
            database.concatenate_group('configurations/cells')
            database.concatenate_group('configurations/pbcs')

            np.testing.assert_allclose(
                database.get_data('configurations/atomic_numbers'),
                np.concatenate([_.get_atomic_numbers() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations/positions'),
                np.concatenate([_.get_positions() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations/cells'),
                np.concatenate([_.get_cell() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations/pbcs'),
                np.concatenate([_.get_pbc() for _ in images]).astype(int)
            )

    def test_add_then_update_with_changes_config(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = HDF5Backend(tmpfile, mode='w')

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            database.insert_data(images)

            for img in images:
                img.info[ATOMS_NAME_FIELD].add('change')
                img.info[ATOMS_LABELS_FIELD].add('another_label')

            database.insert_data(images)

            for n in database.get_data('configurations/names').values():
                assert n[()] == b'change'

            for n in database.get_data('configurations/labels').values():
                assert n[()] == b'another_label'

            for img in images:
                img.info[ATOMS_NAME_FIELD].add('change2')
                img.info[ATOMS_LABELS_FIELD] = {'another_label2'}

            database.insert_data(images)

            for n in database.get_data('configurations/names').values():
                assert n[0] == b'change'
                assert n[1] == b'change2'

            for n in database.get_data('configurations/labels').values():
                assert n[0] == b'another_label'
                assert n[1] == b'another_label2'

            database.get_configuration(next(iter(database['configurations/ids/data'].keys()))).info['_name']

            database.concatenate_group('configurations/atomic_numbers')
            database.concatenate_group('configurations/positions')
            database.concatenate_group('configurations/cells')
            database.concatenate_group('configurations/pbcs')

            np.testing.assert_allclose(
                database.get_data('configurations/atomic_numbers'),
                np.concatenate([_.get_atomic_numbers() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations/positions'),
                np.concatenate([_.get_positions() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations/cells'),
                np.concatenate([_.get_cell() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations/pbcs'),
                np.concatenate([_.get_pbc() for _ in images]).astype(int)
            )


    def test_adding_configurations_no_properties_gen(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = HDF5Backend(tmpfile, mode='w')

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            ids = database.insert_data(images, generator=True)

            list(ids)  # consume generator

            database.concatenate_group('configurations/atomic_numbers')
            database.concatenate_group('configurations/positions')
            database.concatenate_group('configurations/cells')
            database.concatenate_group('configurations/pbcs')

            np.testing.assert_allclose(
                database.get_data('configurations/atomic_numbers'),
                np.concatenate([_.get_atomic_numbers() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations/positions'),
                np.concatenate([_.get_positions() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations/cells'),
                np.concatenate([_.get_cell() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations/pbcs'),
                np.concatenate([_.get_pbc() for _ in images]).astype(int)
            )


    def test_add_configs_props_diff_def(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = HDF5Backend(tmpfile, mode='w')

            returns = build_n(2)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'dft-energy', 'units': 'eV'},
                    'stress': {'field': 'dft-stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'dft-forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},
                }
            }

            for img in images:
                img.info['dft-energy'] = img.info['energy']
                img.info['dft-stress'] = img.info['stress']
                img.arrays['dft-forces'] = img.arrays['forces']

                del img.info['energy']
                del img.info['stress']
                del img.arrays['forces']

            database.insert_data(
                images, property_map=property_map
            )

            database.concatenate_group('properties/default/energy')
            database.concatenate_group('properties/default/stress')
            database.concatenate_group('properties/default/name')
            database.concatenate_group('properties/default/nd-same-shape')

            database.concatenate_group('properties/default/forces')
            database.concatenate_group('properties/default/nd-same-shape-arr')

            with pytest.raises(ConcatenationException):
                database.concatenate_group('properties/default/nd-diff-shapes')

            with pytest.raises(ConcatenationException):
                database.concatenate_group('properties/default/nd-diff-shapes-arr')

            np.testing.assert_allclose(
                database.get_data('properties/default/energy'),
                np.hstack(energies)
            )
            np.testing.assert_allclose(
                database.get_data('properties/default/stress'),
                np.hstack(stress)
            )
            decoded_names = [
                _.decode('utf-8')
                for _ in database.get_data('properties/default/name')
            ]
            assert decoded_names == names
            np.testing.assert_allclose(
                database.get_data('properties/default/nd-same-shape'),
                np.concatenate(nd_same_shape)
            )
            data = database.get_data('properties/default/nd-diff-shapes')
            for a1, a2 in zip(data.values(), nd_diff_shape):
                np.testing.assert_allclose(a1, a2)

            np.testing.assert_allclose(
                database.get_data('properties/default/forces'),
                np.concatenate(forces)
            )
            np.testing.assert_allclose(
                database.get_data('properties/default/nd-same-shape-arr'),
                np.concatenate(nd_same_shape_arr)
            )
            data = database.get_data('properties/default/nd-diff-shapes-arr')
            for a1, a2 in zip(data.values(), nd_diff_shape_arr):
                np.testing.assert_allclose(a1, a2)



    def test_adding_configurations_with_properties_gen(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = HDF5Backend(tmpfile, mode='w')

            returns = build_n(2)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'energy', 'units': 'eV'},
                    'stress': {'field': 'stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},
                }
            }

            ids = database.insert_data(
                images, property_map=property_map, generator=True
            )

            list(ids)  # consume generator

            database.concatenate_group('properties/default/energy')
            database.concatenate_group('properties/default/stress')
            database.concatenate_group('properties/default/name')
            database.concatenate_group('properties/default/nd-same-shape')

            database.concatenate_group('properties/default/forces')
            database.concatenate_group('properties/default/nd-same-shape-arr')

            with pytest.raises(ConcatenationException):
                database.concatenate_group('properties/default/nd-diff-shapes')

            with pytest.raises(ConcatenationException):
                database.concatenate_group('properties/default/nd-diff-shapes-arr')

            np.testing.assert_allclose(
                database.get_data('properties/default/energy'),
                np.hstack(energies)
            )
            np.testing.assert_allclose(
                database.get_data('properties/default/stress'),
                np.hstack(stress)
            )
            decoded_names = [
                _.decode('utf-8')
                for _ in database.get_data('properties/default/name')
            ]
            assert decoded_names == names
            np.testing.assert_allclose(
                database.get_data('properties/default/nd-same-shape'),
                np.concatenate(nd_same_shape)
            )
            data = database.get_data('properties/default/nd-diff-shapes')
            for a1, a2 in zip(data.values(), nd_diff_shape):
                np.testing.assert_allclose(a1, a2)

            np.testing.assert_allclose(
                database.get_data('properties/default/forces'),
                np.concatenate(forces)
            )
            np.testing.assert_allclose(
                database.get_data('properties/default/nd-same-shape-arr'),
                np.concatenate(nd_same_shape_arr)
            )
            data = database.get_data('properties/default/nd-diff-shapes-arr')
            for a1, a2 in zip(data.values(), nd_diff_shape_arr):
                np.testing.assert_allclose(a1, a2)


    def test_add_then_update_with_properties_nochange(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = HDF5Backend(tmpfile, mode='w')

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'energy', 'units': 'eV'},
                    'stress': {'field': 'stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},
                }
            }

            pso = PropertySettings(
                method='VASP',
                description='A basic test calculation',
                files=[('dummy_name', 'dummy file contents')],
            )

            pso_id = database.insert_property_settings(pso)

            database.insert_data(
                images,
                property_map=property_map,
                property_settings={'default': pso_id}
            )

            database.insert_data(
                images,
                property_map=property_map,
                property_settings={'default': pso_id}
            )

            database.concatenate_group('properties/default/energy')
            database.concatenate_group('properties/default/stress')
            database.concatenate_group('properties/default/name')
            database.concatenate_group('properties/default/nd-same-shape')

            database.concatenate_group('properties/default/forces')
            database.concatenate_group('properties/default/nd-same-shape-arr')

            with pytest.raises(ConcatenationException):
                database.concatenate_group('properties/default/nd-diff-shapes')

            with pytest.raises(ConcatenationException):
                database.concatenate_group('properties/default/nd-diff-shapes-arr')

            np.testing.assert_allclose(
                database.get_data('properties/default/energy'),
                np.hstack(energies)
            )
            np.testing.assert_allclose(
                database.get_data('properties/default/stress'),
                np.hstack(stress)
            )
            decoded_names = [
                _.decode('utf-8')
                for _ in database.get_data('properties/default/name')
            ]
            assert decoded_names == names
            np.testing.assert_allclose(
                database.get_data('properties/default/nd-same-shape'),
                np.concatenate(nd_same_shape)
            )
            data = database.get_data('properties/default/nd-diff-shapes')
            for a1, a2 in zip(data.values(), nd_diff_shape):
                np.testing.assert_allclose(a1, a2)

            np.testing.assert_allclose(
                database.get_data('properties/default/forces'),
                np.concatenate(forces)
            )
            np.testing.assert_allclose(
                database.get_data('properties/default/nd-same-shape-arr'),
                np.concatenate(nd_same_shape_arr)
            )
            data = database.get_data('properties/default/nd-diff-shapes-arr')
            for a1, a2 in zip(data.values(), nd_diff_shape_arr):
                np.testing.assert_allclose(a1, a2)


    def test_add_then_update_with_properties_with_change(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = HDF5Backend(tmpfile, mode='w')

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'energy', 'units': 'eV'},
                    'stress': {'field': 'stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},
                }
            }

            pso = PropertySettings(
                method='VASP',
                description='A basic test calculation',
                files=[('dummy_name', 'dummy file contents')],
            )

            pso_id = database.insert_property_settings(pso)

            database.insert_data(
                images,
                property_map=property_map,
                property_settings={'default': pso_id}
            )

            for img in images:
                img.info['energy'] += 100000

            database.insert_data(
                images,
                property_map=property_map,
                property_settings={'default': pso_id}
            )

            for img in images:
                img.info['energy'] += 100000

            database.insert_data(
                images,
                property_map=property_map,
                property_settings={'default': pso_id}
            )

            database.concatenate_group('properties/default/energy')
            database.concatenate_group('properties/default/stress')
            database.concatenate_group('properties/default/name')
            database.concatenate_group('properties/default/nd-same-shape')

            database.concatenate_group('properties/default/forces')
            database.concatenate_group('properties/default/nd-same-shape-arr')

            with pytest.raises(ConcatenationException):
                database.concatenate_group('properties/default/nd-diff-shapes')

            with pytest.raises(ConcatenationException):
                database.concatenate_group('properties/default/nd-diff-shapes-arr')

            np.testing.assert_allclose(
                database.get_data('properties/default/energy'),
                np.concatenate([
                    np.hstack(energies),
                    np.hstack(energies)+100000,
                    np.hstack(energies)+200000
                ])
            )
            np.testing.assert_allclose(
                database.get_data('properties/default/stress'),
                np.concatenate([
                    np.hstack(stress),
                    np.hstack(stress),
                    np.hstack(stress),
                ])
            )
            decoded_names = [
                _.decode('utf-8')
                for _ in database.get_data('properties/default/name')
            ]
            assert decoded_names == names*3
            np.testing.assert_allclose(
                database.get_data('properties/default/nd-same-shape'),
                np.concatenate([
                    np.concatenate(nd_same_shape),
                    np.concatenate(nd_same_shape),
                    np.concatenate(nd_same_shape),
                ])
            )
            data = database.get_data('properties/default/nd-diff-shapes')
            for a1, a2 in zip(data.values(), nd_diff_shape*3):
                np.testing.assert_allclose(a1, a2)

            np.testing.assert_allclose(
                database.get_data('properties/default/forces'),
                np.concatenate([
                    np.concatenate(forces),
                    np.concatenate(forces),
                    np.concatenate(forces),
                ])
            )
            np.testing.assert_allclose(
                database.get_data('properties/default/nd-same-shape-arr'),
                np.concatenate([
                    np.concatenate(nd_same_shape_arr),
                    np.concatenate(nd_same_shape_arr),
                    np.concatenate(nd_same_shape_arr),
                ])
            )
            data = database.get_data('properties/default/nd-diff-shapes-arr')
            for a1, a2 in zip(data.values(), nd_diff_shape_arr*3):
                np.testing.assert_allclose(a1, a2)

    def test_adding_configurations_with_properties(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = HDF5Backend(tmpfile, mode='w')

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'energy', 'units': 'eV'},
                    'stress': {'field': 'stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},
                }
            }

            pso = PropertySettings(
                method='VASP',
                description='A basic test calculation',
                files=[('dummy_name', 'dummy file contents')],
            )

            pso_id = database.insert_property_settings(pso)

            database.insert_data(
                images,
                property_map=property_map,
                property_settings={'default': pso_id}
            )

            database.concatenate_group('properties/default/energy')
            database.concatenate_group('properties/default/stress')
            database.concatenate_group('properties/default/name')
            database.concatenate_group('properties/default/nd-same-shape')

            database.concatenate_group('properties/default/forces')
            database.concatenate_group('properties/default/nd-same-shape-arr')

            with pytest.raises(ConcatenationException):
                database.concatenate_group('properties/default/nd-diff-shapes')

            with pytest.raises(ConcatenationException):
                database.concatenate_group('properties/default/nd-diff-shapes-arr')

            np.testing.assert_allclose(
                database.get_data('properties/default/energy'),
                np.hstack(energies)
            )
            np.testing.assert_allclose(
                database.get_data('properties/default/stress'),
                np.hstack(stress)
            )
            decoded_names = [
                _.decode('utf-8')
                for _ in database.get_data('properties/default/name')
            ]
            assert decoded_names == names
            np.testing.assert_allclose(
                database.get_data('properties/default/nd-same-shape'),
                np.concatenate(nd_same_shape)
            )
            data = database.get_data('properties/default/nd-diff-shapes')
            for a1, a2 in zip(data.values(), nd_diff_shape):
                np.testing.assert_allclose(a1, a2)

            np.testing.assert_allclose(
                database.get_data('properties/default/forces'),
                np.concatenate(forces)
            )
            np.testing.assert_allclose(
                database.get_data('properties/default/nd-same-shape-arr'),
                np.concatenate(nd_same_shape_arr)
            )
            data = database.get_data('properties/default/nd-diff-shapes-arr')
            for a1, a2 in zip(data.values(), nd_diff_shape_arr):
                np.testing.assert_allclose(a1, a2)


    def test_config_add_concat_add_concat(self):

        with tempfile.TemporaryFile() as tmpfile:

            database = HDF5Backend(tmpfile, mode='w')

            images1 = build_n(10)[0]
            atomic_numbers = [atoms.get_atomic_numbers() for atoms in images1]
            positions = [atoms.get_positions() for atoms in images1]
            cells = [atoms.get_cell() for atoms in images1]
            pbcs = [atoms.get_pbc() for atoms in images1]
            database.insert_data(images1)

            database.concatenate_group('configurations/atomic_numbers')

            images2 = build_n(10)[0]
            atomic_numbers += [atoms.get_atomic_numbers() for atoms in images2]
            positions += [atoms.get_positions() for atoms in images2]
            cells += [atoms.get_cell() for atoms in images2]
            pbcs += [atoms.get_pbc() for atoms in images2]
            database.insert_data(images2)

            database.concatenate_group('configurations/atomic_numbers')

            np.testing.assert_allclose(
                database.get_data('configurations/atomic_numbers'),
                np.concatenate(atomic_numbers)
            )
            np.testing.assert_allclose(
                database.get_data(
                    'configurations/positions', concatenate=True, in_memory=True
                ),
                np.concatenate(positions)
            )
            np.testing.assert_allclose(
                database.get_data(
                    'configurations/cells', concatenate=True, in_memory=True
                ),
                np.concatenate(cells)
            )
            np.testing.assert_allclose(
                database.get_data(
                    'configurations/pbcs', concatenate=True, in_memory=True
                ),
                np.concatenate(pbcs)
            )


    def test_get_configurations(self):

        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            images = build_n(10)[0]

            database.insert_data(images)

            # database.concatenate_configurations()

            count = 0
            for atoms, img in zip(database.get_configurations('all'), images):
                assert atoms == img
                count += 1
            assert count == 10


    def test_get_configurations_after_concat(self):

        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            images = build_n(10)[0]

            database.insert_data(images)

            database.concatenate_configurations()

            count = 0
            for atoms, img in zip(database.get_configurations('all'), images):
                assert atoms == img
                count += 1
            assert count == 10


    def test_get_configurations_after_concat_gen(self):

        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            images = build_n(10)[0]

            ids = database.insert_data(images, generator=True)
            list(ids)  # consume generator

            database.concatenate_configurations()

            count = 0
            for atoms, img in zip(
                database.get_configurations('all', generator=True), images
                ):
                assert atoms == img
                count += 1
            assert count == 10


    def test_get_using_returns(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            returns = build_n(10)

            images = returns[0]
            ids    = [_[0] for _ in database.insert_data(images)]

            # database.concatenate_configurations()

            for atoms, img in zip(database.get_configurations(ids), images):
                assert atoms == img


    def test_get_using_returns_gen(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            returns = build_n(10)

            images = returns[0]
            ids    = [
                _[0] for _ in database.insert_data(images, generator=True)
            ]

            # database.concatenate_configurations()

            for atoms, img in zip(
                database.get_configurations(ids, generator=True), images
                ):
                assert atoms == img


class TestPropertyDefinitionsAndSettings:

    def test_invalid_definition(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            property_definition = {
                'property-id': 'this should throw an error',
            }

            with pytest.raises(Exception):
                database.insert_property_definition(property_definition)


    def test_definition_setter_getter(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            property_definition = {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shape': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }

            database.insert_property_definition(property_definition)

            assert database.get_property_definition('default')['definition'] == property_definition


    def test_settings_setter_getter(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            dummy_file_contents = 'this is a dummy file\nwith nonsense contents'

            pso = PropertySettings(
                method='VASP',
                description='A basic test calculation',
                files=[('dummy_name', dummy_file_contents)],
            )

            pso_id = database.insert_property_settings(pso)

            rebuilt_pso = database.get_property_settings(pso_id)['settings']

            assert pso == rebuilt_pso


    def test_settings_duplicate(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            dummy_file_contents = 'this is a dummy file\nwith nonsense contents'

            pso = PropertySettings(
                method='VASP',
                description='A basic test calculation',
                files=[('dummy_name', dummy_file_contents)],
            )

            pso_id = database.insert_property_settings(pso)
            pso_id = database.insert_property_settings(pso)

            rebuilt_pso = database.get_property_settings(pso_id)

            assert pso == rebuilt_pso['settings']

class TestConfigurationSets:

    def test_insert_cs(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            images = build_n(10)[0]

            ids = [_[0] for _ in database.insert_data(images)]

            cs_id = database.insert_configuration_set(
                ids, description='A basic configuration set'
            )

            desc = database[f'configuration_sets/{cs_id}'].attrs['description']

            assert desc == 'A basic configuration set'

            rebuilt_ids = database[f'configuration_sets/{cs_id}'].attrs['configuration_ids'][()].tolist()

            # rebuilt_ids = database.get_data(
            #     f'configuration_sets/{cs_id}/ids',
            #     ravel=True, in_memory=True, as_str=True
            # ).tolist()

            assert rebuilt_ids == ids

            for im1, im2 in zip(
                database.get_configurations(rebuilt_ids), images
                ):
                assert im1 == im2


class TestDatasets:

    def test_insert_ds(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = HDF5Backend(tmpfile, mode='w')

            images = build_n(10)[0]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'energy', 'units': 'eV'},
                    'stress': {'field': 'stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},
                }
            }

            ids = database.insert_data(
                images, property_map=property_map
            )

            co_ids1, pr_ids1 = list(zip(*ids))

            cs_id1 = database.insert_configuration_set(
                co_ids1, description='A basic configuration set'
            )

            images = build_n(10)[0]

            for img in images:
                img.info['energy'] += 100000

            ids = database.insert_data(
                images, property_map=property_map
            )

            co_ids2, pr_ids2 = list(zip(*ids))

            cs_id2 = database.insert_configuration_set(
                co_ids2, description='A basic configuration set'
            )

            ds_id = database.insert_dataset(
                cs_ids=[cs_id1, cs_id2],
                pr_ids=pr_ids1+pr_ids2,
                authors='colabfit',
                links='https://colabfit.openkim.org/',
                description='An example dataset'
            )

            assert database[f'datasets/{ds_id}'].attrs['authors'] == ['colabfit']


from datetime import datetime
from hashlib import sha512

import dateutil.parser
#from pyspark.sql import functions as sf
#from pyspark.sql.types import StringType

#from colabfit.tools.schema import configuration_set_schema
from colabfit.tools.utilities import ELEMENT_MAP, _empty_dict_from_schema


class ConfigurationSet:
    """
    A configuration set defines a group of configurations and aggregates
    information about those configurations to improve queries.

    Note that a configuration set should only be constructed by loading from an
    existing database (in order to aggregate the info)

    Attributes:

        configuration_ids (list):
            A list of all attached configuration IDs

        description (str):
            A human-readable description of the configuration set

        aggregated_info (dict):
            A dictionary of information that was aggregated from all of the
            attached configurations. Contains the following information:

                * nconfigurations: the total number of configurations
                * nsites: the total number of sites
                * nelements: the total number of unique element types
                * elements: the element types
                * individual_elements_ratios: a set of elements ratios generated by
                  looping over each configuration, extracting its concentration of each
                  element, and adding the tuple of concentrations to the set
                * total_elements_ratios: the ratio of the total count of atoms of each
                  element type over nsites
                * chemical_formula_reduced: the set of all reduced chemical formulae
                * chemical_formula_anonymous: the set of all anonymous chemical formulae
                * chemical_formula_hill: the set of all hill chemical formulae
                * nperiodic_dimensions: the set of all numbers of periodic dimensions
                * dimension_types: the set of all periodic boundary choices

    """

    def __init__(self, config_df, name, description, dataset_id, ordered=False):
        self.name = name
        self.description = description
        self.dataset_id = dataset_id
        # self.ordered = ordered
        self.spark_row = self.to_spark_row(config_df)
        self.id = f"CS_{self.name}_{self.dataset_id}"
        self._hash = hash(self)
        self.spark_row["id"] = self.id
        self.spark_row["extended_id"] = f"{self.name}__{self.id}"
        self.spark_row["hash"] = str(self._hash)

    def to_spark_row(self, config_df):
        row_dict = _empty_dict_from_schema(configuration_set_schema)
        row_dict["name"] = self.name
        row_dict["description"] = self.description
        row_dict["nconfigurations"] = config_df.select("id").distinct().count()
        row_dict["last_modified"] = dateutil.parser.parse(
            datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        )
        # config_df = config_df.withColumn(
        #     "nsites_multiple", sf.col("nsites") * sf.col("multiplicity")
        # )
        row_dict["nsites"] = config_df.agg({"nsites": "sum"}).first()[0]
        row_dict["elements"] = sorted(
            config_df.withColumn("exploded_elements", sf.explode("elements"))
            .agg(sf.collect_set("exploded_elements").alias("exploded_elements"))
            .select("exploded_elements")
            .take(1)[0][0]
        )
        row_dict["nperiodic_dimensions"] = config_df.agg(
            sf.collect_set("nperiodic_dimensions")
        ).collect()[0][0]
        row_dict["dimension_types"] = config_df.agg(
            sf.collect_set("dimension_types")
        ).collect()[0][0]
        atomic_ratios_df = (
            config_df.select("atomic_numbers")
            #     config_df.select("atomic_numbers", "multiplicity")
            #     .withColumn(
            #         "repeated_numbers",
            #         sf.expr(
            #            "transform(atomic_numbers, x -> array_repeat(x, multiplicity))"
            #         ),
            #     )
            # .withColumn("single_element", sf.explode(sf.flatten("repeated_numbers")))
            .withColumn("single_element", sf.explode("atomic_numbers"))
        )
        total_elements = atomic_ratios_df.count()
        print(total_elements, row_dict["nsites"])
        assert total_elements == row_dict["nsites"]
        atomic_ratios_df = atomic_ratios_df.groupBy("single_element").count()
        atomic_ratios_df = atomic_ratios_df.withColumn(
            "ratio", sf.col("count") / total_elements
        )
        atomic_ratios_coll = (
            atomic_ratios_df.withColumn(
                "element",
                sf.udf(lambda x: ELEMENT_MAP[int(x)], StringType())(
                    sf.col("single_element")
                ),
            )
            .select("element", "ratio")
            .collect()
        )
        row_dict["total_elements_ratios"] = [
            x[1] for x in sorted(atomic_ratios_coll, key=lambda x: x["element"])
        ]
        row_dict["nelements"] = len(row_dict["elements"])
        row_dict["nsites"] = config_df.agg({"nsites": "sum"}).first()[0]
        row_dict["dataset_id"] = self.dataset_id

        return row_dict

    def __hash__(self):
        cs_hash = sha512()
        cs_hash.update(self.id.encode("utf-8"))
        return int(cs_hash.hexdigest(), 16)

    def __str__(self):
        return "ConfigurationSet(description='{}', nconfigurations={})".format(
            self.description,
            self.spark_row["nconfigurations"],
        )

    def __repr__(self):
        return str(self)



from ase import Atoms
from colabfit.tools.configuration import AtomicConfiguration
from colabfit.tools.utilities import convert_stress
from pathlib import Path
import re


def mlip_cfg_reader(symbol_map, filepath):
    with open(filepath, "rt") as f:
        energy = None
        forces = None
        coords = []
        cell = []
        symbols = []
        config_count = 0
        for line in f:
            if line.strip().startswith("Size"):
                size = int(f.readline().strip())
            elif line.strip().lower().startswith("supercell"):
                cell.append([float(x) for x in f.readline().strip().split()])
                cell.append([float(x) for x in f.readline().strip().split()])
                cell.append([float(x) for x in f.readline().strip().split()])
            elif line.strip().startswith("Energy"):
                energy = float(f.readline().strip())
            elif line.strip().startswith("PlusStress"):
                stress_keys = line.strip().split()[-6:]
                stress = [float(x) for x in f.readline().strip().split()]
                stress = convert_stress(stress_keys, stress)
            elif line.strip().startswith("AtomData:"):
                keys = line.strip().split()[1:]
                if "fx" in keys:
                    forces = []
                for i in range(size):
                    li = {
                        key: val for key, val in zip(keys, f.readline().strip().split())
                    }
                    symbols.append(symbol_map[li["type"]])
                    if "cartes_x" in keys:
                        coords.append(
                            [
                                float(c)
                                for c in [
                                    li["cartes_x"],
                                    li["cartes_y"],
                                    li["cartes_z"],
                                ]
                            ]
                        )
                    elif "direct_x" in keys:
                        coords.append(
                            [
                                float(c)
                                for c in [
                                    li["direct_x"],
                                    li["direct_y"],
                                    li["direct_z"],
                                ]
                            ]
                        )
                    if "fx" in keys:
                        forces.append(
                            [float(f) for f in [li["fx"], li["fy"], li["fz"]]]
                        )
            elif line.startswith("END_CFG"):
                if "cartes_x" in keys:
                    config = Atoms(positions=coords, symbols=symbols, cell=cell)
                elif "direct_x" in keys:
                    config = Atoms(scaled_positions=coords, symbols=symbols, cell=cell)
                config.info["energy"] = energy
                if forces:
                    config.info["forces"] = forces
                config.info["stress"] = stress  # Stress units appear to be kbar (?)
                config.info["_name"] = f"{filepath.stem}__index__{config_count}"
                config_count += 1
                yield AtomicConfiguration.from_ase(config)
                forces = None
                stress = []
                coords = []
                cell = []
                symbols = []
                energy = None


##############################################################
# VASP OUTCAR parser
##############################################################


vasp_coord_regex = re.compile(
    r"^\s+(?P<x>\-?\d+\.\d+)\s+(?P<y>\-?\d+\.\d+)\s+(?P<z>\-?\d+\.\d+)\s+(?P<fx>\-?"
    r"\d+\.\d+)\s+(?P<fy>\-?\d+\.\d+)\s+(?P<fz>\-?\d+\.\d+)"
)
param_re = re.compile(
    r"[\s+]?(?P<param>[A-Z_]+)(\s+)?=(\s+)?(?P<val>-?([\d\w\.\-]+)?\.?)"
    r"[\s;]?(?P<unit>eV)?\:?"
)
IGNORE_PARAMS = [
    "VRHFIN",
    "LEXCH",
    "EATOM",
    "TITEL",
    "LULTRA",
    "IUNSCR",
    "RPACOR",
    "POMASS",
    "RCORE",
    "RWIGS",
    "ENMAX",
    "RCLOC",
    "LCOR",
    "LPAW",
    "EAUG",
    "DEXC",
    "RMAX",
    "RAUG",
    "RDEP",
    "RDEPT",
]


def vasp_contcar_parser(fp):
    with open(fp, "r") as f:
        for i in range(5):
            _ = f.readline()
        line = f.readline()
        symbols = line.strip().split()
        counts = [int(x) for x in f.readline().strip().split()]
        symbol_arr = []
        for symbol, count in zip(symbols, counts):
            symbol_arr.extend([symbol] * count)
        return symbol_arr


def config_namer_by_filepath(fp, dataset_path):
    ds_fp_str = "__".join(dataset_path.absolute().parts).replace("/", "")
    name = "__".join(fp.absolute().parts[:-1]).replace("/", "")
    name = name.replace(ds_fp_str + "__", "")
    return name


def vasp_outcar_reader(symbols, fp):
    with open(fp, "r") as f:
        incar = dict()
        cinput = dict()
        in_latt = False
        in_coords = False
        lattice = []
        pos = []
        forces = []
        stress = None
        potcars = set()
        energy = None
        for line in f:
            # Prelim handling
            if line.strip() == "":
                pass

            # handle lattice
            elif "direct lattice vectors" in line:
                in_latt = True
                lattice = []
                pass
            elif in_latt is True:
                latt = line.strip().replace("-", " -").split()
                lattice.append([float(x) for x in [latt[0], latt[1], latt[2]]])
                if len(lattice) == 3:
                    in_latt = False
            elif "POTCAR:" in line:
                potcars.add(" ".join(line.strip().split()[1:]))
            elif "FREE ENERGIE OF THE ION-ELECTRON SYSTEM" in line:
                _ = f.readline()
                _, _, _, _, energy, units = f.readline().strip().split()
                if len(pos) > 0:
                    cinput["incar"] = incar
                    config = Atoms(positions=pos, symbols=symbols, cell=lattice)
                    config.info["input"] = cinput
                    config.info["input"]["potcars"] = list(potcars)
                    # config.info["outcar"] = outcar
                    if stress is not None:
                        config.info["stress"] = stress
                    config.info["forces"] = forces
                    config.info["energy"] = float(energy)
                    yield config
                    forces = []
                    stress = None
                    pos = []
                    energy = None
            elif "POSITION" in line:
                in_coords = True
                pass
            elif in_coords is True:
                if "--------" in line:
                    pass
                elif "total drift" in line:
                    in_coords = False
                    if energy is not None:
                        cinput["incar"] = incar
                        config = Atoms(positions=pos, symbols=symbols, cell=lattice)
                        config.info["input"] = cinput
                        config.info["input"]["potcars"] = list(potcars)
                        # config.info["outcar"] = outcar
                        config.info["forces"] = forces
                        if stress is not None:
                            config.info["stress"] = stress
                        config.info["energy"] = float(energy)
                        yield config
                        forces = []
                        stress = None
                        pos = []
                        energy = None
                    else:
                        pass
                else:
                    cmatch = vasp_coord_regex.search(line)
                    pos.append(
                        [float(p) for p in [cmatch["x"], cmatch["y"], cmatch["z"]]]
                    )
                    forces.append(
                        [float(p) for p in [cmatch["fx"], cmatch["fy"], cmatch["fz"]]]
                    )
            elif "Direction" in line and "XX" in line:
                stress_keys = [x.lower() for x in line.strip().split()[1:]]
            elif "in kB" in line:
                stress = [float(x) for x in line.strip().split()[2:]]
                stress = convert_stress(stress_keys, stress)

            else:
                pass
                # print("something went wrong")


def vasp_get_kpoints(fp):
    with open(fp, "r") as f:
        # f.readline() # if skipping first line
        kpoints = "".join(f.readlines())
    return kpoints


def vasp_parse_incar(fp):
    with open(fp, "r") as f:
        return f.read()


def file_finder(fp, file_glob, count=0):
    if count > 5:
        return None
    elif next(fp.glob(file_glob), None) is not None:
        # file_glob in [f.name for f in fp.glob("*")]:
        return next(fp.glob(file_glob))
    else:
        count += 1
        return file_finder(fp.parent, file_glob, count)


def vasp_outcar_wrapper(data_dir: Path, dataset_path, CO_METADATA=None):
    outcars = sorted(list(data_dir.rglob("OUTCAR")))
    for filepath in outcars:
        name = config_namer_by_filepath(filepath, dataset_path)
        poscar = next(filepath.parent.glob(filepath.name.replace("OUTCAR", "POSCAR")))
        symbols = vasp_contcar_parser(poscar)
        kpoints_file = file_finder(filepath.parent, "KPOINTS")
        kpoints = vasp_get_kpoints(kpoints_file)
        incar_file = file_finder(filepath.parent, "INCAR")
        incar = vasp_parse_incar(incar_file)

        config_gen = vasp_outcar_reader(fp=filepath, symbols=symbols)
        for i, config in enumerate(config_gen):
            config.info["name"] = f"{name}_{i}"
            config.info["input"]["kpoints"] = kpoints
            config.info["input"]["incar"] = incar
            yield AtomicConfiguration.from_ase(config, CO_METADATA)



"""
Each schema for VastDB has a corresponding schema for a dataframe with
the same fields, but with non-stringified lists.
config_schema (with stringified lists), for example, has a corresponding
config_df_schema (with non-stringified lists).
"""

from pyspark.sql.types import (
    ArrayType,
    BooleanType,
    DoubleType,
    IntegerType,
    LongType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)

from colabfit.tools.utilities import get_stringified_schema

NSITES_COL_SPLITS = 20

config_df_schema = StructType(
    [
        StructField("id", StringType(), True),
        StructField("hash", StringType(), True),
        StructField("last_modified", TimestampType(), True),
        StructField("dataset_ids", ArrayType(StringType()), True),
        StructField("configuration_set_ids", ArrayType(StringType()), True),
        StructField("chemical_formula_hill", StringType(), True),
        StructField("chemical_formula_reduced", StringType(), True),
        StructField("chemical_formula_anonymous", StringType(), True),
        StructField("elements", ArrayType(StringType()), True),
        StructField("elements_ratios", ArrayType(DoubleType()), True),
        StructField("atomic_numbers", ArrayType(IntegerType()), True),
        StructField("nsites", IntegerType(), True),
        StructField("nelements", IntegerType(), True),
        StructField("nperiodic_dimensions", IntegerType(), True),
        StructField("cell", ArrayType(ArrayType(DoubleType())), True),
        StructField("dimension_types", ArrayType(IntegerType()), True),
        StructField("pbc", ArrayType(BooleanType()), True),
        StructField("names", ArrayType(StringType()), True),
        StructField("labels", ArrayType(StringType()), True),
        StructField(f"positions", ArrayType(ArrayType(DoubleType())), True),
        #StructField("metadata_id", StringType(), True),
        #StructField("metadata_path", StringType(), True),
        #StructField("metadata_size", IntegerType(), True),
    ]
    #+ [
    #    StructField(f"positions_{i:02d}", ArrayType(ArrayType(DoubleType())), True)
    #    for i in range(NSITES_COL_SPLITS)
    #]
)
config_schema = get_stringified_schema(config_df_schema)
config_md_schema = config_df_schema.add(StructField("metadata", StringType(), True))


property_object_df_schema = StructType(
    [
        StructField("id", StringType(), True),
        StructField("hash", StringType(), True),
        StructField("last_modified", TimestampType(), True),
        StructField("configuration_id", StringType(), True),
        StructField("dataset_id", StringType(), True),
        StructField("multiplicity", IntegerType(), True),
        StructField("metadata_id", StringType(), True),
        StructField("metadata_path", StringType(), True),
        StructField("metadata_size", IntegerType(), True),
        StructField("software", StringType(), True),
        StructField("method", StringType(), True),
        StructField("chemical_formula_hill", StringType(), True),
        StructField("energy", DoubleType(), True),
    ]
    + [
        StructField(f"atomic_forces_{i:02d}", ArrayType(ArrayType(DoubleType())), True)
        for i in range(NSITES_COL_SPLITS)
    ]
    + [
        StructField("cauchy_stress", ArrayType(ArrayType(DoubleType())), True),
        StructField("cauchy_stress_volume_normalized", BooleanType(), True),
        StructField("electronic_band_gap", DoubleType(), True),
        StructField("electronic_band_gap_type", StringType(), True),
        StructField("formation_energy", DoubleType(), True),
        StructField("adsorption_energy", DoubleType(), True),
        StructField("atomization_energy", DoubleType(), True),
    ]
    # TODO: Add schema associated with new properties: selection/descriptor
)

property_object_schema = get_stringified_schema(property_object_df_schema)
property_object_md_schema = property_object_df_schema.add(
    StructField("metadata", StringType(), True)
)


dataset_df_schema = StructType(
    [
        StructField("id", StringType(), True),
        StructField("hash", StringType(), True),
        StructField("name", StringType(), True),
        StructField("last_modified", TimestampType(), True),
        StructField("nconfigurations", IntegerType(), True),
        StructField("nproperty_objects", LongType(), True),
        StructField("nsites", LongType(), True),
        StructField("nelements", IntegerType(), True),
        StructField("elements", ArrayType(StringType()), True),
        StructField("total_elements_ratios", ArrayType(DoubleType()), True),
        StructField("nperiodic_dimensions", ArrayType(IntegerType()), True),
        StructField("dimension_types", ArrayType(ArrayType(IntegerType())), True),
        StructField("energy_count", LongType(), True),
        StructField("energy_mean", DoubleType(), True),
        StructField("energy_variance", DoubleType(), True),
        StructField("atomization_energy_count", LongType(), True),
        StructField("adsorption_energy_count", LongType(), True),
        StructField("formation_energy_count", LongType(), True),
        StructField("atomic_forces_count", LongType(), True),
        StructField("electronic_band_gap_count", LongType(), True),
        StructField("cauchy_stress_count", LongType(), True),
        StructField("authors", ArrayType(StringType()), True),
        StructField("description", StringType(), True),
        StructField("extended_id", StringType(), True),
        StructField("license", StringType(), True),
        StructField("links", StringType(), True),
        StructField("publication_year", StringType(), True),
        StructField("doi", StringType(), True),
        # StructField("labels", ArrayType(StringType()), True),
    ]
)

dataset_schema = get_stringified_schema(dataset_df_schema)


configuration_set_df_schema = StructType(
    [
        StructField("id", StringType(), True),
        StructField("hash", StringType(), True),
        StructField("last_modified", TimestampType(), True),
        StructField("nconfigurations", IntegerType(), True),
        StructField("nperiodic_dimensions", ArrayType(IntegerType()), True),
        StructField("dimension_types", ArrayType(ArrayType(IntegerType())), True),
        StructField("nsites", LongType(), True),
        StructField("nelements", IntegerType(), True),
        StructField("elements", ArrayType(StringType()), True),
        StructField("total_elements_ratios", ArrayType(DoubleType()), True),
        StructField("description", StringType(), True),
        StructField("name", StringType(), True),
        StructField("dataset_id", StringType(), True),
        StructField("ordered", BooleanType(), True),
        StructField("extended_id", StringType(), True),
    ]
)

configuration_set_schema = get_stringified_schema(configuration_set_df_schema)

co_cs_mapping_schema = StructType(
    [
        StructField("configuration_id", StringType(), True),
        StructField("configuration_set_id", StringType(), True),
    ]
)



energy_pd = {
    "property-id": "tag:staff@noreply.colabfit.org,2024-06-01:property/energy",
    "property-name": "energy",
    "property-title": "Energy from which atomic forces calculations have been derived",
    "property-description": (
        "The calculated energy property used to derive the value of the atomic forces "
        "property. Values are converted if necessary to units of eV, and normalized to "
        "per-configuration energy (i.e., energy reported in this column is not energy "
        "per atom)."
    ),
    "energy": {
        "type": "float",
        "has-unit": True,
        "extent": [],
        "required": True,
        "description": "The potential energy of the system.",
    },
    "per-atom": {
        "type": "bool",
        "has-unit": False,
        "extent": [],
        "required": False,
        "description": 'If True, "energy" is the total energy of the system divided by '
        "the number of atoms in the configuration.",
    },
    "reference-energy": {
        "type": "float",
        "has-unit": True,
        "extent": [],
        "required": False,
        "description": 'If provided, then "energy" is the total energy (either '
        "of the whole system, or per-atom) "
        "minus the energy of a reference configuration (E = E_0 - E_reference). "
        'Note that "reference-energy" is just provided for documentation, and that '
        '"energy" should already have this value subtracted off. The reference '
        'energy must have the same units as "energy".',
    },
}

atomic_forces_pd = {
    "property-id": "tag:staff@noreply.colabfit.org,2022-05-30:property/atomic-forces",
    "property-name": "atomic-forces",
    "property-title": "Atomic forces from a static calculation",
    "property-description": (
        "The x, y, and z components of the force on each atom, calculated as a "
        "molecule’s energy differentiated with respect to atom positions and "
        "representing the negative gradient of the energy with respect to "
        "atom positions."
    ),
    "forces": {
        "type": "float",
        "has-unit": True,
        "extent": [":", 3],
        "required": True,
        "description": "The [x,y,z] components of the force on each particle.",
    },
}

cauchy_stress_pd = {
    "property-id": "tag:staff@noreply.colabfit.org,2022-05-30:property/cauchy-stress",
    "property-name": "cauchy-stress",
    "property-title": "Cauchy stress tensor from a static calculation",
    "property-description": "Full 3x3 Cauchy stress tensor from a calculation of a static configuration.",
    "stress": {
        "type": "float",
        "has-unit": True,
        "extent": [3, 3],
        "required": True,
        "description": "Cauchy stress tensor.",
    },
    "volume-normalized": {
        "type": "bool",
        "has-unit": False,
        "extent": [],
        "required": False,
        "description": "If True, the stress has been multiplied by the cell volume.",
    },
}

atomization_energy_pd = {
    "property-id": "tag:staff@noreply.colabfit.org,2022-11-18:property/atomization-energy",
    "property-name": "atomization-energy",
    "property-title": "Atomization energy from a static calculation",
    "property-description": "Formation energy from a calculation of a static "
    "configuration which uses atomic systems to compute reference energies. Energies "
    "must be specified to be per-atom if true. If the reference energies are known, "
    "these must be specified as well.",
    "energy": {
        "type": "float",
        "has-unit": True,
        "extent": [],
        "required": False,
        "description": "The atomization energy of the system.",
    },
    "per-atom": {
        "type": "bool",
        "has-unit": False,
        "extent": [],
        "required": False,
        "description": 'If True, "energy" is the atomization energy of the system '
        "divided by the number of atoms in the configuration.",
    },
    "reference-energy": {
        "type": "float",
        "has-unit": True,
        "extent": [],
        "required": False,
        "description": "If known, represents values which have been subtracted from "
        "the base energy, resulting in the atomization energy.",
    },
}

adsorption_energy_pd = {
    "property-id": "tag:staff@noreply.colabfit.org,2022-11-18:property/adsorption-energy",
    "property-name": "adsorption-energy",
    "property-title": "Adsorption energy from a static calculation",
    "property-description": "Adsorption energy from a calculation of a static "
    "configuration which uses non-atomic systems to compute reference energies. "
    "Energies must be specified to be per-atom if true. If the reference energies are "
    "known, this must be specified as well.",
    "energy": {
        "type": "float",
        "has-unit": True,
        "extent": [],
        "required": False,
        "description": "The adsorption energy of the system.",
    },
    "per-atom": {
        "type": "bool",
        "has-unit": False,
        "extent": [],
        "required": False,
        "description": 'If True, "energy" is the adsorption energy of the system '
        "divided by the number of atoms in the configuration.",
    },
    "reference-energy": {
        "type": "float",
        "has-unit": True,
        "extent": [],
        "required": False,
        "description": "If known, represents values which have been subtracted from "
        "the base energy, resulting in the adsorption energy.",
    },
}


formation_energy_pd = {
    "property-id": "tag:staff@noreply.colabfit.org,2022-11-18:property/formation-energy",
    "property-name": "formation-energy",
    "property-title": "Formation energy from a static calculation",
    "property-description": "Formation energy from a calculation of a static "
    "configuration which uses non-atomic systems to compute reference energies. "
    "Energies must be specified to be per-atom if true. If the reference energies are "
    "known, this must be specified as well.",
    "energy": {
        "type": "float",
        "has-unit": True,
        "extent": [],
        "required": False,
        "description": "The formation energy of the system.",
    },
    "per-atom": {
        "type": "bool",
        "has-unit": False,
        "extent": [],
        "required": False,
        "description": 'If True, "energy" is the formation energy of the system '
        "divided by the number of atoms in the configuration.",
    },
    "reference-energy": {
        "type": "float",
        "has-unit": True,
        "extent": [],
        "required": False,
        "description": "If known, represents values which have been subtracted from "
        "the base energy, resulting in the formation energy.",
    },
}


band_gap_pd = {
    "property-id": "tag:staff@noreply.colabfit.org,2022-11-18:property/band-gap",
    "property-name": "band-gap",
    "property-title": "Band gap energy from a static calculation",
    "property-description": "Band gap energy from a calculation of a static "
    "configuration.",
    "energy": {
        "type": "float",
        "has-unit": True,
        "extent": [],
        "required": False,
        "description": "The band gap energy of the system.",
    },
    "type": {
        "type": "string",
        "has-unit": False,
        "extent": [],
        "required": True,
        "description": "The type of band gap calculation: [direct, indirect, unknown].",
    },
}

quests_descriptor_pd = {
            "property-id": "tag:staff@noreply.colabfit.org,2024-12-09:property/quests-descriptor",
            "property-name": 'quests-descriptor',
            "property-title": "QUESTS descriptor",
            "property-description": "The concatenation of a list of sorted"
            "neighbor distances and average triplet bond lengths",

            "descriptors": {  # example: (N,D) arrays
                "type": "float",
                "has-unit": False,
                "extent": [":", ":"],
                "required": True,
                "description": "The per-atom descriptors. N is equal to"
                "the number of atoms and D is (2*num-nearest-neighbors)-1",
            },
            "num-nearest-neighbors": {  # example: 32
                "type": "int",
                "has-unit": False,
                "extent": [],
                "required": True,
                "description": "The number of nearest neighbors "
                "included in the calculation. Determines the dimensionality "
                "of the quests descriptor: (2*num-nearest-neighbors)-1",
            },
            "cutoff": {  # example: 5.0
                "type": "float",
                "has-unit": True,
                "extent": [],
                "required": True,
                "description": "The cutoff distance in calculation",
            },
        }

custom_mask_pd = {
        'property-id': 'tag:staff@noreply.colabfit.org,2024-12-09: property/selection-mask',
        'property-name': 'selection',
        'property-title': 'Selection mask',
        'property-description': 'List of bools determining if atom is selected or not',
        'mask': {
            'type': 'bool',
            'has-unit': False,
            'extent': [':'],
            'required': True,
            'description': 'The per-atom selection'
            }
        }




import json
import os
import sys
from ast import literal_eval
from hashlib import sha512

import numpy as np
#import pyarrow as pa
from pyspark.sql import Row
from pyspark.sql import functions as sf
from pyspark.sql.types import (
    BooleanType,
    DoubleType,
    IntegerType,
    LongType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)


def _format_for_hash(v):
    if isinstance(v, np.ndarray):
        if np.issubdtype(v.dtype, np.floating):
            return np.round_(v.astype(np.float64), decimals=16)
        elif np.issubdtype(v.dtype, np.integer):
            return v.astype(np.int64)
        elif np.issubdtype(v.dtype, bool):
            return v.astype(np.int64)
        else:
            return v
    elif isinstance(v, list):
        return np.array(v).data.tobytes()
    elif isinstance(v, dict):
        return str(v).encode("utf-8")
    elif isinstance(v, str):
        return v.encode("utf-8")
    elif isinstance(v, (int, float)):
        return np.array(v).data.tobytes()
    elif isinstance(v, tuple):
        return np.array(v).data.tobytes()
    else:
        return v


def _hash(row, identifying_key_list, include_keys_in_hash=False):
    identifying_key_list = sorted(identifying_key_list)
    identifiers = [row[i] for i in identifying_key_list]
    # sort_for_hash = ["positions", "atomic_forces"]
    _hash = sha512()
    for k, v in zip(identifying_key_list, identifiers):
        if v is None or v == "[]":
            continue
        # elif k in sort_for_hash:
        #     v = np.array(v)
        #     sorted_v = v[np.lexsort((
        #         v[:, 2],
        #         v[:, 1],
        #         v[:, 0],
        #     ))]
        #     _hash.update(bytes(_format_for_hash(sorted_v)))
        else:
            if include_keys_in_hash:
                _hash.update(bytes(_format_for_hash(k)))
            _hash.update(bytes(_format_for_hash(v)))
    return int(_hash.hexdigest(), 16)


def get_spark_field_type(schema, field_name):
    for field in schema:
        if field.name == field_name:
            return field.dataType
    raise ValueError(f"Field name {field_name} not found in schema")


def get_stringified_schema(schema):
    new_fields = []
    for field in schema:
        if field.dataType.typeName() == "array":
            new_fields.append(StructField(field.name, StringType(), field.nullable))
        else:
            new_fields.append(field)
    return StructType(new_fields)


def spark_to_arrow_type(spark_type):
    """
    Convert PySpark type to PyArrow type.
    Do not include field.nullable, as this conflicts with vastdb-sdk
    """
    if isinstance(spark_type, IntegerType):
        return pa.int32()
    elif isinstance(spark_type, LongType):
        return pa.int64()
    elif isinstance(spark_type, DoubleType):
        return pa.float64()
    elif isinstance(spark_type, StringType):
        return pa.string()
    elif isinstance(spark_type, TimestampType):
        return pa.timestamp("us")
    elif isinstance(spark_type, BooleanType):
        return pa.bool_()
    elif isinstance(spark_type, StructType):
        return pa.schema(
            [
                pa.field(field.name, spark_to_arrow_type(field.dataType))
                for field in spark_type
            ]
        )
    else:
        raise ValueError(f"Unsupported type: {spark_type}")


def spark_schema_to_arrow_schema(spark_schema):
    """
    Convert PySpark schema to a PyArrow Schema.
    """
    fields = []
    for field in spark_schema:
        if field.name == "$row_id":
            fields.append(pa.field(field.name, pa.uint64()))
        else:
            fields.append(pa.field(field.name, spark_to_arrow_type(field.dataType)))
    return pa.schema(fields)


def arrow_record_batch_to_rdd(schema, batch):
    names = schema.fieldNames()
    arrays = [batch.column(i) for i in range(batch.num_columns)]
    for i in range(batch.num_rows):
        yield {names[j]: array[i].as_py() for j, array in enumerate(arrays)}


def _empty_dict_from_schema(schema):
    empty_dict = {}
    for field in schema:
        empty_dict[field.name] = None
    return empty_dict


def _sort_dict(dictionary):
    keys = list(dictionary.keys())
    keys.sort()
    return {k: dictionary[k] for k in keys}


def _parse_unstructured_metadata(md_json):
    if md_json == {}:
        return {
            "metadata": None,
            "metadata_id": None,
            "metadata_path": None,
            "metadata_size": None,
        }
    md = {}
    for key, val in md_json.items():
        if key in ["_id", "hash", "colabfit-id", "last_modified", "software", "method"]:
            continue
        if isinstance(val, dict):
            if "source-value" in val.keys():
                val = val["source-value"]
        if isinstance(val, list) and len(val) == 1:
            val = val[0]
        if isinstance(val, np.ndarray):
            val = val.tolist()
        if isinstance(val, dict):
            val = _sort_dict(val)
        if isinstance(val, bytes):
            val = val.decode("utf-8")
        md[key] = val
    md = _sort_dict(md)
    md_hash = str(_hash(md, md.keys(), include_keys_in_hash=True))
    md["hash"] = md_hash
    md["id"] = f"MD_{md_hash[:25]}"
    split = md["id"][-4:]
    filename = f"{md['id']}.json"
    after_bucket = os.path.join(split, filename)
    metadata = json.dumps(md)
    return {
        "metadata": metadata,
        "metadata_id": md["id"],
        "metadata_path": after_bucket,
        "metadata_size": sys.getsizeof(metadata),
    }


# if not os.path.isfile(full_path):
#         # Write iff the ID is new and unique
#         os.makedirs(os.path.dirname(full_path), exist_ok=True)
#         with open(full_path, "w") as f:
#             json.dump(md, f)


def stringify_lists(row_dict):
    """
    Replace list/tuple fields with comma-separated strings.
    Spark and Vast both support array columns, but the connector does not,
    so keeping cell values in list format crashes the table.
    Use with dicts
    TODO: Remove when no longer necessary
    """
    for key, val in row_dict.items():
        if isinstance(val, (list, tuple, dict)):
            row_dict[key] = str(val)
        # Below would convert numpy arrays to comma-separated
        elif isinstance(val, np.ndarray):
            row_dict[key] = str(val.tolist())
    return row_dict


def stringify_rows(row):
    """
    Convert list/tuple fields to comma-separated strings.
    Use with spark Rows
    Should be mapped as DataFrame.rdd.map(stringify_rows)"""
    row_dict = row.asDict()
    for key, val in row_dict.items():
        if isinstance(val, (list, tuple, dict)):
            row_dict[key] = str(val)
    new_row = Row(**row_dict)
    return new_row


def stringify_rows_to_dict(row):
    """
    Convert list/tuple fields to comma-separated strings.
    Use with spark Rows
    Should be mapped as DataFrame.rdd.map(stringify_rows)"""
    row_dict = row.asDict()
    for key, val in row_dict.items():
        if isinstance(val, (list, tuple, dict)):
            row_dict[key] = str(val)
    return row_dict


def stringify_row_dict(row_dict):
    for key, val in row_dict.items():
        if isinstance(val, (list, tuple, dict)):
            row_dict[key] = str(val)
    return row_dict


def unstringify(row):
    """Should be mapped as DataFrame.rdd.map(unstringify)"""
    row_dict = row.asDict()
    try:
        for key, val in row_dict.items():
            if key == "metadata":
                continue
            elif isinstance(val, str) and len(val) > 0 and val[0] in ["["]:
                dval = literal_eval(row[key])
                row_dict[key] = dval
            else:
                continue
        new_row = Row(**row_dict)
    except Exception as e:
        print(e)
        print(row_dict)
    return new_row


def unstring_df_val(val):
    if isinstance(val, str) and len(val) > 0 and val[0] in ["["]:
        dval = literal_eval(val)
        return dval
    else:
        return val


def stringify_df_val(val):
    if isinstance(val, (list, tuple, dict)):
        return str(val)
    else:
        return val


def unstringify_row_dict(row_dict):
    """Should be mapped as rdd.map(unstringify_row_dict)"""
    for key, val in row_dict.items():
        if key == "metadata":
            continue
        if isinstance(val, str) and len(val) > 0 and val[0] in ["{", "["]:
            dval = literal_eval(row_dict[key])
            row_dict[key] = dval
    return row_dict


def append_ith_element_to_rdd_labels(row_elem):
    """
    row_elem: tuple created by joining two RDD.zipWithIndex
    new_labels: list of labels
    """
    (index, (co_row, new_labels)) = row_elem
    val = co_row.get("labels")
    if val is None:
        val = new_labels
    else:
        val.extend(new_labels)
        val = list(set(val))
    co_row["labels"] = val
    return co_row


def add_elem_to_row_dict(col, elem, row_dict):
    val = row_dict.get(col)
    if val is None:
        val = [elem]
    else:
        val.append(elem)
        val = list(set(val))
    row_dict[col] = val
    return row_dict


def convert_stress(keys, stress):
    """Convert a size-6 array of stress components to a 3x3 matrix

    In particular for VASP output. Assumes symmetric matrix.
    Check order of keys."""
    stresses = {k: s for k, s in zip(keys, stress)}
    return [
        [stresses["xx"], stresses["xy"], stresses["xz"]],
        [stresses["xy"], stresses["yy"], stresses["yz"]],
        [stresses["xz"], stresses["yz"], stresses["zz"]],
    ]


##########################################################
# Functions for splitting oversize arrays to columns
##########################################################
def get_max_string_length(df, column_name):

    max_len = (
        df.select(column_name)
        .select(sf.length(column_name).alias("string_length"))
        .agg(sf.max("string_length"))
        .collect()[0][0]
    )
    if max_len is None:
        return 0
    return max_len


def split_long_string_cols(df, column_name: str, max_string_length: int):
    """
    Splits a long string column into multiple columns based on a maximum string length.
    :param df: Input DataFrame with array cols already stringified
    :param column_name: Name of the column containing the long string
    :param max_string_length: Maximum length for each split string
    :return: DataFrame with the long string split across multiple columns
    """
    overflow_columns = [
        f"{'_'.join(column_name.split('_')[:-1])}_{i + 1:02}" for i in range(19)
    ]
    if not all([col in df.columns for col in overflow_columns]):
        raise ValueError("Overflow columns not found in target DataFrame schema")
    if get_max_string_length(df, column_name) <= max_string_length:
        for col in overflow_columns:
            df = df.withColumn(col, sf.lit("[]").cast(StringType()))
        return df
    print(f"Column split: {column_name}")
    all_columns = [column_name] + overflow_columns
    tmp_columns = [f"{col_name}_tmp" for col_name in all_columns]
    df = df.withColumn("total_length", sf.length(sf.col(column_name)))
    max_string_length = 35000 // 15
    substring_exprs = [
        sf.when(
            sf.length(sf.col(column_name)) - (i * max_string_length) > 0,
            sf.substring(
                sf.col(column_name), (i * max_string_length + 1), max_string_length
            ),
        )
        .otherwise(sf.lit("[]"))
        .alias(col_name)
        for i, col_name in enumerate(tmp_columns)
    ]
    df = df.select("*", *substring_exprs)
    for tmp_col, col in zip(tmp_columns, all_columns):
        df = df.drop(col).withColumnRenamed(f"{tmp_col}", col)
    df = df.drop("total_length")
    return df


# ##########################################################
# # Functions for writing values to files
# ##########################################################


# def _write_value(path_prefix, id_str, filetype, BUCKET_DIR, value):
#     """i.e.: _write_value(
#     value=co['positions'],
#     'CO/positions', co['id'],
#     'txt', '/save/here'
#     )
#     """
#     # Use the final 4 digits of the id for an ~1000-way split
#     split = id_str[-4:]
#     filename = f"{id_str}.{filetype}"
#     full_path = Path(BUCKET_DIR) / path_prefix / split / filename
#     full_path.parent.mkdir(parents=True, exist_ok=True)
#     full_path.write_text(str(value))
#     return full_path


# def write_value_to_file(path_prefix, extension, BUCKET_DIR, write_column, row):
#     """i.e.: partial(_write_value(
#     'CO/positions',
#     'txt',
#     '/save/here'
#     'positions',
#     )
#     """
#     id = row["id"]
#     # Use the final 4 digits of the id for an ~1000-way split
#     value = row[write_column]
#     split = id[-4:]
#     filename = f"{id}.{extension}"
#     full_path = Path(BUCKET_DIR) / path_prefix / split / filename
#     full_path.parent.mkdir(parents=True, exist_ok=True)
#     full_path.write_text(str(value))
#     row_dict = row.asDict()
#     row_dict[write_column] = str(full_path)
#     return Row(**row_dict)


# def multi_value_to_file(path_prefixes, extension, BUCKET_DIR, write_columns, row):
#     """i.e.: partial(_write_value(
#     'CO/positions',
#     'txt',
#     '/save/here'
#     'positions',
#     )
#     """
#     id = row["id"]
#     split = id[-4:]

#     row_dict = row.asDict()
#     for write_column, path_prefix in zip(write_columns, path_prefixes):
#         value = row[write_column]
#         filename = f"{id}.{extension}"
#         full_path = Path(BUCKET_DIR) / path_prefix / split / filename
#         full_path.parent.mkdir(parents=True, exist_ok=True)
#         full_path.write_text(str(value))
#         row_dict[write_column] = str(full_path)
#     return Row(**row_dict)


ELEMENT_MAP = {
    1: "H",
    2: "He",
    3: "Li",
    4: "Be",
    5: "B",
    6: "C",
    7: "N",
    8: "O",
    9: "F",
    10: "Ne",
    11: "Na",
    12: "Mg",
    13: "Al",
    14: "Si",
    15: "P",
    16: "S",
    17: "Cl",
    18: "Ar",
    19: "K",
    20: "Ca",
    21: "Sc",
    22: "Ti",
    23: "V",
    24: "Cr",
    25: "Mn",
    26: "Fe",
    27: "Co",
    28: "Ni",
    29: "Cu",
    30: "Zn",
    31: "Ga",
    32: "Ge",
    33: "As",
    34: "Se",
    35: "Br",
    36: "Kr",
    37: "Rb",
    38: "Sr",
    39: "Y",
    40: "Zr",
    41: "Nb",
    42: "Mo",
    43: "Tc",
    44: "Ru",
    45: "Rh",
    46: "Pd",
    47: "Ag",
    48: "Cd",
    49: "In",
    50: "Sn",
    51: "Sb",
    52: "Te",
    53: "I",
    54: "Xe",
    55: "Cs",
    56: "Ba",
    57: "La",
    58: "Ce",
    59: "Pr",
    60: "Nd",
    61: "Pm",
    62: "Sm",
    63: "Eu",
    64: "Gd",
    65: "Tb",
    66: "Dy",
    67: "Ho",
    68: "Er",
    69: "Tm",
    70: "Yb",
    71: "Lu",
    72: "Hf",
    73: "Ta",
    74: "W",
    75: "Re",
    76: "Os",
    77: "Ir",
    78: "Pt",
    79: "Au",
    80: "Hg",
    81: "Tl",
    82: "Pb",
    83: "Bi",
    84: "Po",
    85: "At",
    86: "Rn",
    87: "Fr",
    88: "Ra",
    89: "Ac",
    90: "Th",
    91: "Pa",
    92: "U",
    93: "Np",
    94: "Pu",
    95: "Am",
    96: "Cm",
    97: "Bk",
    98: "Cf",
    99: "Es",
    100: "Fm",
    101: "Md",
    102: "No",
    103: "Lr",
    104: "Rf",
    105: "Db",
    106: "Sg",
    107: "Bh",
    108: "Hs",
    109: "Mt",
    110: "Ds",
    111: "Rg",
    112: "Cn",
    113: "Nh",
    114: "Fl",
    115: "Mc",
    116: "Lv",
    117: "Ts",
    118: "Og",
}



import h5py
import json
import warnings
import datetime
import numpy as np
from tqdm import tqdm
from ase import Atoms
from copy import deepcopy
from hashlib import sha512

from kim_property.definition import check_property_definition
from kim_property.definition import PROPERTY_ID as VALID_KIM_ID

from colabfit import (
    HASH_SHIFT,
    ATOMS_LABELS_FIELD,
    ATOMS_LAST_MODIFIED_FIELD,
    ATOMS_NAME_FIELD,
    STRING_DTYPE_SPECIFIER
)
from colabfit.tools.configuration import Configuration
from colabfit.tools.property import Property
from colabfit.tools.property_settings import PropertySettings

class HDF5Backend(h5py.File):
    """
    A HDF5Backend extends a PyTables (HDF5) file, but provides additional
    functionality for construction, filtering, exploring, ...

    In general, a HDF5Backend works by grouping the data (coordinates, lattice
    vectors, property fields, ...), maintaining columns of IDs to identify which
    data corresponds to each Configuration/Property/etc., and providing
    additional columns of IDs for mapping from the data to any linked entries
    (configurations, properties, configuration sets, or property settings).

    The underlying HDF5 filesystem has the following structure:

    C  = number of configurations
    I  = number of fields in :attr:`Configuration.info`
    A  = number of fields in :attr:`Configuration.arrays`

    P  = number of properties defined for the database
    Fp = number of fields for a given property

    S  = number of property settings definitions in the database

    G  = number of configuration sets in the database
    D  = number of datasets in the database

    /root
        /configurations
            This group stores all of the information necessary to construct each
            of the configurations in the database.

            /ids
                /data
                    An array of all Configuration IDs in the database

            /names
                Same as /atomic_numbers, but for configuration names. Note that
                a single configuration may have been given multiple names.

            /labels
                Same as /atomic_numbers, but for labels

            /last_modified
                Same as /atomic_numbers, but for datetime strings specifying
                when the configuration was last modified.

            /atomic_numbers
                .attrs
                    concatenated
                        A boolean indicating if /data has been concatenated or
                        not.
                /data
                    The atomic numbers for all configurations in the database.

                    If :code:`concatenated==True`, then /data will contain
                    multiple arrays, one per configuration.

                    If :code:`concatenated==True`, then /data will contain
                    a single array that was generated using something like:

                    ..code-block:: python

                        np.concatenate(
                            # /data/<co_id> for co_id in /data
                        )
                /slices
                        A group where /slices/config_id_1 specifies how to slice
                        /data in order to get the data corresponding to
                        config_id_1.

                        The data can always be extracted using something like:

                        ..code-block:: python

                            slices = database['configurations/field/slices'][()]

                            database.get_data(
                                'configurations/atomic_numbers'
                            )[()][slices]

                        If :code:`concatenated==False`, then /slices/config_id_1
                        will just be config_id_1.

                        If :code:`concatenated==True`, then /slices/config_id_1
                        will be an integer array specifying rows of /data
            /positions
                Same as /atomic_numbers, but for atomic positions
            /cells
                Same as /atomic_numbers, but for lattice vectors
            /pbcs
                Same as /atomic_numbers, but for periodic boundary conditions
            /constraints
                This group is not supported yet. It will be added in the future,
                and will have a similar structure to /atomic_numbers
        /properties
            This group stores all of the computed properties.

            /ids
                As in /configurations/ids, but for the Property IDs

            /types
                As in /ids, but the type of the property (e.g., "property_1")

            /configuration_ids
                A list of tuples of Configuration IDs specifying the
                Configurations associated with each Property. Uses the
                /data, /slices form.

            /settings_ids
                A list of Property Settings IDs. Uses the /data, /slices
                form.

            /last_modified
                Same as configurations/atomic_numbers, but for datetime strings
                specifying when the configuration was last modified.

            /property_1
                .attrs
                    definition:
                        An OpenKIM Property Definition as a (serialized)
                        dictionary

                    last_modified:
                        A datetime string specifying when the object was
                        modified last.

                /field_1
                    /data
                        As in /configurations/info/info_field_1/data
                    /slices
                        As in /configurations/info/info_field_1/slices, but
                        mapping to the Property IDs

                .
                .
                .
                /field_Fp
            .
            .
            .
            /property_P
        /property_settings
            The property settings.

            /property_settings_id_1
                .attrs
                    method: VASP, QuantumEspresso, ...

                    description:
                        A description of the calculation

                    last_modified:
                        A datetime string specifying when the object was
                        modified last.

                    labels:
                        Labels applied to the property settings

                /files
                    Any files associated with the property settings, stored as
                    raw text strings. Uses the /data form, as in
                    /configurations/ids.
            .
            .
            .
            /property_settings_id_S
        /configuration_sets
            The configuration sets defining groups of configurations.

            /configuration_set_id_1
                .attrs
                    description:
                        Human-readable description of the set

                    configuration_ids:
                        The list of configuration IDs in the configuration set

                    last_modified:
                        A datetime string specifying when the object was
                        modified last.
                /ids
                    A list of Configuration IDs that belong to the configuration
                    set. Useful for indexing /configurations fields and
                    /properties/property fields. Uses the /data form, as in
                    /configurations/ids.
            .
            .
            .
            /configuration_set_id_G
        /datasets
            /dataset_1
                .attrs
                    authors:
                        A list of author names

                    links:
                        A list of external links (e.g., journal articles, Git
                        repos, ...)

                    description:
                        A human-readable description of the dataset

                    configuration_set_ids:
                        The list of configuration set IDs

                    property_ids:
                        The list of property IDs

                    last_modified:
                        A datetime string specifying when the object was
                        modified last.
    """

    def __init__(self, name, mode='r', **kwargs):
        """
        Args:
            name (str):
                The path to the database file

            mode (str):
                'r', 'w', 'a', or 'w+'
        """

        super().__init__(name=name, mode=mode, **kwargs)

        # Build all the base groups
        if 'configurations' not in self:
            g = self.create_group('configurations')

            for sub_group_name in [
                'ids', 'atomic_numbers', 'positions', 'cells', 'pbcs',
                'names', 'labels', 'last_modified'
                ]:
                sub_g = g.create_group(sub_group_name)
                sub_g.attrs['concatenated'] = False
                sub_g.create_group('data', track_order=True)

                if sub_group_name != 'ids':
                    sub_g.create_group('slices')

        if 'properties' not in self:
            g = self.create_group('properties')

            g_ids = g.create_group('ids')
            g_ids.attrs['concatenated'] = False
            g_ids.create_group('data', track_order=True)

            g_types = g.create_group('types')
            g_types.attrs['concatenated'] = False
            g_types.create_group('data', track_order=True)

            g_settings = g.create_group('configuration_ids')
            g_settings.attrs['concatenated'] = False
            g_settings.create_group('data', track_order=True)

            g_settings = g.create_group('settings_ids')
            g_settings.attrs['concatenated'] = False
            g_settings.create_group('data', track_order=True)

        for n in  ['property_settings', 'configuration_sets', 'datasets']:
            if n not in self:
                self.create_group(n)


    def insert_data(
        self,
        configurations, property_map=None, property_settings=None,
        generator=False, verbose=False
        ):
        """
        Inserts the configurations into the databas, and any specified
        properties, and returns an iterable of 2-tuples, where the first entry
        in each tuple is the ID of the inserted property, and the second entry
        is the ID of the associated configuration.

        A configuration is added to the database by extracting seven fields from
        the Configuration object:

            * atomic numbers (:meth:`get_atomic_numbers`)
            * atomic positions (:meth:`get_positions`)
            * cell lattice vectors (:meth:`get_cell`)
            * cell periodic boundary conditions (:meth:`get_pbc`)
            * names (:attr:`info['_name'])
            * labels (:attr:`info['_labels'])
            * last_modified (:meth:`datetime.now`)

        A property is added to the database by extracting the fields specified
        by the property definition off of the configuration. Note that an error
        will be thrown if the property hasn't been defined yet using
        :meth:`~colabfit.tools.database.HDF5Backend.add_property_definition`.

        Set :code:`generator=True` when the configurations
        can't all fit in memory at the same time. NOTE: if `generator==True`,
        then the configurations will only be added to the dataset once the
        returned IDs have been iterated over

        Note that new properties can be added to existing configurations by
        passing the same configurations to insert_data(), but using a new
        property map.

        Adding two different configurations with the same property can be used
        to add one property attached to multiple configurations.

        Example:

        ..code-block:: python

            database.add_property_definition(...)

            property_map = {
                'property-name-1': {
                    'property-field-1': {
                        'field': 'ase-field-1',
                        'units': 'units-1',
                    }
                }
            }

            pso_id_1 = database.add_property_settings(
                PropertySettings(...)
            )

            property_settings = {
                'property-name': pso_id_1
            }

            database.insert_data(
                configurations, property_map, property_settings
            )


        Args:

            configurations (list or Configuration):
                The list of configurations to be added.

            property_map (dict):
                A dictionary that is used to specify how to load a defined
                property off of a configuration. Note that the top-level keys in
                the map must be the names of properties that have been
                previously defined using
                :meth:`~colabfit.tools.database.HDF5Backend.add_property_definition`.

                If None, only loads the configuration information (atomic
                numbers, positions, lattice vectors, and periodic boundary
                conditions).

            property_settings (dict)
                key = property name (same as top-level keys in property_map).
                val = property settings ID that has been previously entered into
                the database using
                :meth:`~colabfit.tools.database.HDF5Backend.insert_property_settings`

            generator (bool):
                If true, this function becomes a generator which only adds the
                configurations one at a time. This is useful if the
                configurations can't all fit in memory at the same time. Default
                is False.

            verbose (bool):
                If True, prints progress bar

        Returns:

            additions (iterable):
                An iterable of 2-tuples, where the first entry in each tuple is
                the ID of the inserted property, and the second entry is the ID
                of the associated configuration.
        """

        if isinstance(configurations, Configuration):
            configurations = [configurations]

        if property_map is None:
            property_map = {}

        if property_settings is None:
            property_settings = {}

        for settings_id in property_settings.values():
            if settings_id not in self['property_settings']:
                raise MissingEntryError(
                    "The property settings object with ID '{}' does"\
                    " not exist in the database".format(settings_id)
                )

        property_definitions = {
            pname: self.get_property_definition(pname)['definition']
            for pname in property_map
        }

        if generator:
            return self._insert_data_gen(
                configurations,
                property_definitions, property_map, property_settings,
                verbose=verbose
            )
        else:
            return self._insert_data(
                configurations,
                property_definitions, property_map, property_settings,
                verbose=verbose
            )


    # @profile
    def _insert_data_gen(
        self, configurations,
        property_definitions, property_map, property_settings,
        verbose=False
        ):
        if isinstance(configurations, Configuration):
            configurations = [configurations]

        ignore_keys = {
            'property-id', 'property-title', 'property-description',
            'last_modified', 'definition'
        }
        expected_keys = {
            pname: set(
                property_map[pname][f]['field']
                for f in property_definitions[pname].keys() - ignore_keys
                # property_definitions[pname].keys()
            )
            for pname in property_map
        }

        for ai, atoms in enumerate(tqdm(
            configurations,
            desc='Inserting data in HDF5 backend',
            disable=not verbose
            )):
            config_id = str(hash(atoms))

            g = self['configurations/ids/data']

            if config_id in g:  # Identical CO already exists in the dataset
                # So just update /names, /labels, and /last_modified
                for f2 in ['names', 'labels', 'last_modified']:
                    g = self[f'configurations/{f2}']
                    if g.attrs['concatenated']:
                        raise ConcatenationException(
                            "Trying to update a configuration after "\
                            "concatenating is not allowed."
                        )

                # Now append to existing datasets
                # Names
                data = self[f'configurations/names/data/{config_id}']
                new_names = atoms.info[ATOMS_NAME_FIELD]
                names_set = set(new_names) - set(data.asstr()[()])
                data.resize((data.shape[0]+len(names_set),) + data.shape[1:])
                data[-len(names_set):] = np.array(
                    list(names_set), dtype=STRING_DTYPE_SPECIFIER
                )

                # Labels
                data = self[f'configurations/labels/data/{config_id}']
                labels = atoms.info[ATOMS_LABELS_FIELD]
                data.resize(
                    (data.shape[0]+len(labels),) + data.shape[1:]
                )
                data[-len(labels):] = list(labels)

                # Last modified
                data = self[f'configurations/names/data/{config_id}']
                data = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')

            else:
                # Adding a new CO
                # Names
                g = self['configurations/names']
                names = list(
                    atoms.info[ATOMS_NAME_FIELD]
                )
                g['data'].create_dataset(
                    name=config_id,
                    shape=(len(names),),
                    maxshape=(None,),
                    dtype=STRING_DTYPE_SPECIFIER,
                    data=names
                )
                g[f'slices/{config_id}'] = np.array(
                    config_id, dtype=STRING_DTYPE_SPECIFIER
                )
                # Labels
                g = self['configurations/labels']
                labels = list(atoms.info[ATOMS_LABELS_FIELD])
                g['data'].create_dataset(
                    name=config_id,
                    shape=(len(labels),),
                    maxshape=(None,),
                    dtype=STRING_DTYPE_SPECIFIER,
                    data=labels
                )
                g[f'slices/{config_id}'] = np.array(
                    config_id, dtype=STRING_DTYPE_SPECIFIER
                )
                # Last modified
                now = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
                self[f'configurations/last_modified/data/{config_id}'] = now
                self[f'configurations/last_modified/slices/{config_id}'] = config_id

                # Save the ID
                g = self['configurations/ids/data']
                g.create_dataset(
                    name=config_id,
                    shape=1,
                    data=np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)
                )

                # Save all fundamental information about the configuration
                g = self['configurations/atomic_numbers']
                g['data'].create_dataset(
                    name=config_id,
                    data=atoms.get_atomic_numbers(),
                )
                g[f'slices/{config_id}'] = np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)

                g = self['configurations/positions']
                g['data'].create_dataset(
                    name=config_id,
                    data=atoms.get_positions()
                )
                g[f'slices/{config_id}'] = np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)

                g = self['configurations/cells']
                g['data'].create_dataset(
                    name=config_id,
                    data=np.array(atoms.get_cell())
                )
                g[f'slices/{config_id}'] = np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)

                g = self['configurations/pbcs']
                g['data'].create_dataset(
                    name=config_id,
                    data=atoms.get_pbc().astype(int),
                )
                g[f'slices/{config_id}'] = np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)

            # Try to load all of the specified properties
            available_keys = set().union(atoms.info.keys(), atoms.arrays.keys())

            # Flag for tracking if we need to return (None, config_id)
            returned_something = False

            for pname, pmap in property_map.items():

                # Pre-check to avoid having to delete partially-added properties
                missing_keys = expected_keys[pname] - available_keys
                if missing_keys:
                    warnings.warn(
                        "Configuration {} is missing keys {} during "\
                        "insert_data. Available keys: {}. Skipping".format(
                            ai, missing_keys, available_keys
                        )
                    )
                    continue

                prop = Property.from_definition(
                    pname, property_definitions[pname],
                    atoms, pmap
                )

                prop_id = str(hash(prop))

                # Check for duplicate property. Note that if even a single field
                # is changed, it is considered a new property. This can lead to
                # duplicate data (for the unchanged fields), but is still the
                # desired behaviour.
                if prop_id in self['properties/ids/data']:
                    yield (config_id, prop_id)
                    continue

                # Add the data; group should already exist
                for field in property_definitions[pname].keys()-ignore_keys:
                    g = self[f'properties/{pname}/{field}']

                    # Try to convert field into either a float or string array
                    v = prop[field]['source-value']

                    if isinstance(v, str):
                        v = np.atleast_1d(v).astype(STRING_DTYPE_SPECIFIER)
                    if isinstance(v, set):
                        # These should always be sets of strings
                        v = np.atleast_1d(list(v)).astype(STRING_DTYPE_SPECIFIER)
                    else:
                        v = np.atleast_1d(v)

                    g['data'].create_dataset(
                        name=prop_id,
                        data=v
                    )
                    g[f'slices/{prop_id}'] = np.array(
                        prop_id, dtype=STRING_DTYPE_SPECIFIER
                    )

                g = self['properties/ids/data']
                g.create_dataset(
                    name=prop_id,
                    shape=1,
                    data=np.array(prop_id, dtype=STRING_DTYPE_SPECIFIER)
                )

                g = self[f'properties/types/data/{prop_id}'] = pname

                g = self[f'properties/configuration_ids']
                g['data'].create_dataset(
                    name=prop_id,
                    shape=1,
                    data=np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)
                )
                g[f'slices/{prop_id}'] = np.array(
                    prop_id, dtype=STRING_DTYPE_SPECIFIER
                )

                # Attach property settings, if any were given
                if pname in property_settings:
                    settings_id = property_settings[pname]

                    g = self[f'properties/settings_ids']
                    g['data'].create_dataset(
                        name=prop_id,
                        shape=1,
                        data=np.array(settings_id, dtype=STRING_DTYPE_SPECIFIER)
                    )
                    g[f'slices/{prop_id}'] = np.array(
                        prop_id, dtype=STRING_DTYPE_SPECIFIER
                    )

                now = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
                self[f'properties/last_modified/data/{prop_id}'] = now
                self[f'properties/last_modified/slices/{prop_id}'] = prop_id

                # yield (prop_id, config_id)
                yield (config_id, prop_id)
                returned_something = True

            if not returned_something:
                yield (config_id, None)


    def _insert_data(
        self, configurations,
        property_definitions, property_map, property_settings, verbose
        ):
        if isinstance(configurations, Configuration):
            configurations = [configurations]

        ignore_keys = {
            'property-id', 'property-title', 'property-description',
            'last_modified', 'definition'
        }
        expected_keys = {
            pname: set(
                property_map[pname][f]['field']
                for f in property_definitions[pname].keys() - ignore_keys
                # property_definitions[pname].keys()
            )
            for pname in property_map
        }

        additions = []

        for ai, atoms in enumerate(tqdm(
            configurations,
            desc='Inserting data',
            disable=not verbose
            )):
            config_id = str(hash(atoms))

            g = self['configurations/ids/data']

            if config_id in g:  # Identical CO already exists in the dataset
                # So just update /names, /labels, and /last_modified
                for f2 in ['names', 'labels', 'last_modified']:
                    g = self[f'configurations/{f2}']
                    if g.attrs['concatenated']:
                        raise ConcatenationException(
                            "Trying to update a configuration after "\
                            "concatenating is not allowed."
                        )

                # Now append to existing datasets
                # Names
                data = self[f'configurations/names/data/{config_id}']
                new_names = atoms.info[ATOMS_NAME_FIELD]
                names_set = set(new_names) - set(data.asstr()[()])
                data.resize((data.shape[0]+len(names_set),) + data.shape[1:])
                data[-len(names_set):] = np.array(
                    list(names_set), dtype=STRING_DTYPE_SPECIFIER
                )

                # Labels
                data = self[f'configurations/labels/data/{config_id}']
                labels = atoms.info[ATOMS_LABELS_FIELD]
                data.resize(
                    (data.shape[0]+len(labels),) + data.shape[1:]
                )
                data[-len(labels):] = list(labels)

                # Last modified
                data = self[f'configurations/names/data/{config_id}']
                data = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')

            else:
                # Adding a new CO
                # Names
                g = self['configurations/names']
                names = list(
                    atoms.info[ATOMS_NAME_FIELD]
                )
                g['data'].create_dataset(
                    name=config_id,
                    shape=(len(names),),
                    maxshape=(None,),
                    dtype=STRING_DTYPE_SPECIFIER,
                    data=names
                )
                g[f'slices/{config_id}'] = np.array(
                    config_id, dtype=STRING_DTYPE_SPECIFIER
                )
                # Labels
                g = self['configurations/labels']
                labels = list(atoms.info[ATOMS_LABELS_FIELD])
                g['data'].create_dataset(
                    name=config_id,
                    shape=(len(labels),),
                    maxshape=(None,),
                    dtype=STRING_DTYPE_SPECIFIER,
                    data=labels
                )
                g[f'slices/{config_id}'] = np.array(
                    config_id, dtype=STRING_DTYPE_SPECIFIER
                )
                # Last modified
                now = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
                self[f'configurations/last_modified/data/{config_id}'] = now
                self[f'configurations/last_modified/slices/{config_id}'] = config_id

                # Save the ID
                g = self['configurations/ids/data']
                g.create_dataset(
                    name=config_id,
                    shape=1,
                    data=np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)
                )

                # Save all fundamental information about the configuration
                g = self['configurations/atomic_numbers']
                g['data'].create_dataset(
                    name=config_id,
                    data=atoms.get_atomic_numbers(),
                )
                g[f'slices/{config_id}'] = np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)

                g = self['configurations/positions']
                g['data'].create_dataset(
                    name=config_id,
                    data=atoms.get_positions()
                )
                g[f'slices/{config_id}'] = np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)

                g = self['configurations/cells']
                g['data'].create_dataset(
                    name=config_id,
                    data=np.array(atoms.get_cell())
                )
                g[f'slices/{config_id}'] = np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)

                g = self['configurations/pbcs']
                g['data'].create_dataset(
                    name=config_id,
                    data=atoms.get_pbc().astype(int),
                )
                g[f'slices/{config_id}'] = np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)

            # Try to load all of the specified properties
            available_keys = set().union(atoms.info.keys(), atoms.arrays.keys())

            # Flag for tracking if we need to return (None, config_id)
            returned_something = False

            for pname, pmap in property_map.items():


                # Pre-check to avoid having to delete partially-added properties
                missing_keys = expected_keys[pname] - available_keys
                if missing_keys:
                    warnings.warn(
                        "Configuration {} is missing keys {} during "\
                        "insert_data. Available keys: {}. Skipping".format(
                            ai, missing_keys, available_keys
                        )
                    )
                    continue

                prop = Property.from_definition(
                    pname, property_definitions[pname],
                    atoms, pmap
                )

                prop_id = str(hash(prop))

                # Check for duplicate property. Note that if even a single field
                # is changed, it is considered a new property. This can lead to
                # duplicate data (for the unchanged fields), but is still the
                # desired behaviour.
                if prop_id in self['properties/ids/data']:
                    additions.append((config_id, prop_id))
                    continue

                # Add the data; group should already exist
                for field in property_definitions[pname].keys()-ignore_keys:
                    g = self[f'properties/{pname}/{field}']

                    # Try to convert field into either a float or string array
                    v = prop[field]['source-value']

                    if isinstance(v, str):
                        v = np.atleast_1d(v).astype(STRING_DTYPE_SPECIFIER)
                    if isinstance(v, set):
                        # These should always be sets of strings
                        v = np.atleast_1d(list(v)).astype(STRING_DTYPE_SPECIFIER)
                    else:
                        v = np.atleast_1d(v)

                    g['data'].create_dataset(
                        name=prop_id,
                        data=v
                    )
                    g[f'slices/{prop_id}'] = np.array(
                        prop_id, dtype=STRING_DTYPE_SPECIFIER
                    )

                g = self['properties/ids/data']
                g.create_dataset(
                    name=prop_id,
                    shape=1,
                    data=np.array(prop_id, dtype=STRING_DTYPE_SPECIFIER)
                )

                g = self[f'properties/types/data/{prop_id}'] = pname

                g = self[f'properties/configuration_ids']
                g['data'].create_dataset(
                    name=prop_id,
                    shape=1,
                    data=np.array(config_id, dtype=STRING_DTYPE_SPECIFIER)
                )
                g[f'slices/{prop_id}'] = np.array(
                    prop_id, dtype=STRING_DTYPE_SPECIFIER
                )

                # Attach property settings, if any were given
                if pname in property_settings:
                    settings_id = property_settings[pname]

                    g = self[f'properties/settings_ids']
                    g['data'].create_dataset(
                        name=prop_id,
                        shape=1,
                        data=np.array(settings_id, dtype=STRING_DTYPE_SPECIFIER)
                    )
                    g[f'slices/{prop_id}'] = np.array(
                        prop_id, dtype=STRING_DTYPE_SPECIFIER
                    )

                now = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
                self[f'properties/last_modified/data/{prop_id}'] = now
                self[f'properties/last_modified/slices/{prop_id}'] = prop_id

                # yield (prop_id, config_id)
                additions.append((config_id, prop_id))
                returned_something = True

            if not returned_something:
                additions.append((config_id, None))

        return additions


    def concatenate_group(self, group, chunks=None):
        """
        Attempt to concatenate all of the datasets in a group. Raise an
        exception if the datasets in the group have incompatible shapes.

        Args:

            group_name (str or group):
                The name of a group in the database, or the group object

            chunks (tuple):
                An optional argument describing how to chunk the concatenated
                array. Chunk shapes should be chosed based on the desired access
                pattern. See `chunked storage <https://docs.h5py.org/en/stable/high/dataset.html#chunked-storage>_
                in the h5py documentation for more details.
        """

        if isinstance(group, str):
            group = self[group]

        if 'data' not in group:
            raise InvalidGroupError(
                "The group '{}' does not have a 'data' sub-group".format(
                    group
                )
            )

        # Just initialize an array of the right size
        try:
            example = next(iter(group['data'].values()))
        except StopIteration:
            raise ConcatenationException(
                "Group '{}/data' sub-group is empty. Make sure to consume the "\
                "generator from insert_data() if you used "\
                "generator=True.".format(group.name)
            )

        n = sum((_.shape[0] for _ in group['data'].values()))


        bigshape = (n,) + example.shape[1:]
        data = group.create_dataset(
            name=None, shape=bigshape, dtype=example.dtype, chunks=chunks
        )

        start = 0
        problem_adding = False

        if '_root_concatenated' in group['data']:
            # Copy any already-concatenated data
            ds = group['data/_root_concatenated']
            stop = start + ds.shape[0]
            data[start:stop] = ds
            start += ds.shape[0]

        for ds_name, ds in group['data'].items():
            if ds_name == '_root_concatenated':
                # This was added first
                continue

            try:
                # Copy the data over
                stop = start + ds.shape[0]
                data[start:stop] = ds

                if 'slices' in group:
                    # Some groups (like IDs) don't need to be sliced
                    key = ds.name.split("/")[-1]

                    del group[f'slices/{key}']
                    group[f'slices/{key}'] = np.arange(start, stop)

                start += ds.shape[0]
            except Exception as e:
                problem_adding = True
                break

        if problem_adding:
            # Get rid of the new dataset that was added
            del data
            raise ConcatenationException(
                "Expected shape {}, but got shape {} for dataset '{}'".format(
                    (None,)+bigshape[1:], ds.shape, ds.name
                )
            )
        else:
            # Get rid of all of the old stuff
            del group['data']
            g = group.create_group('data', track_order=True)
            g['_root_concatenated'] = data
            group.attrs['concatenated'] = True


    def get_data(
        self, group, ids=None, in_memory=False,
        concatenate=False, ravel=False, as_str=False
        ):
        """
        Returns all of the datasets in the 'data' sub-group of
        :code:`<group_name>`.

        Args:

            group_name (str or group):
                The name of a group in the database, or the group object

            ids (list):
                The list of IDs to return the data for. If None, returns the
                data for the entire group.

            in_memory (bool):
                If True, converts each of the datasets to a Numpy array before
                returning.

            concatenate (bool):
                If True, concatenates the data before returning. Only available
                if :code:`in_memory==True`.

            ravel (bool):
                If True, concatenates and ravels the data before returning. Only
                available if :code:`in_memory==True`.

            as_str (bool):
                If True, tries to call :code:`asstr()` to convert from an HDF5
                bytes array to an array of strings
        """

        if isinstance(group, str):
            group = self[group]

        if isinstance(ids, str):
            ids = [ids]

        if concatenate or ravel:
            if not in_memory:
                raise ConcatenationException(
                    "Cannot use concatenate=True or ravel=True without "\
                    "in_memory=True"
                )

        if 'data' not in group:
            raise InvalidGroupError(
                "The group '{}' does not have a 'data' sub-group".format(
                    group
                )
            )
        else:
            g = group['data']

            if group.attrs['concatenated']:
                data = g['_root_concatenated']

                if ids:
                    slices = np.concatenate(
                        group[f'slices/{k}'][()] for k in ids
                    )
                    data = data[slices]

                if as_str:
                    data = data.asstr()
                if in_memory:
                    data = data[()]
                if ravel:
                    data = data.ravel()

                return data
            else:
                keys = ids if ids is not None else g.keys()
                data = [g[k] for k in keys]

                if as_str:
                    data = [_.asstr() for _ in data]

                if concatenate:
                    return np.concatenate([_[()] for _ in data])

                if ravel:
                    return np.concatenate([_[()] for _ in data]).ravel()

                return {
                    # encode since /slices will have bytes
                    k.encode('utf-8'): ds for k, ds in zip(keys, data)
                }


    def get_configuration(self, i):
        """
        Returns a single configuration by calling :meth:`get_configurations`
        """
        return self.get_configurations([i])[0]

    def get_configurations(self, ids, generator=False, verbose=False):
        """
        A generator that returns in-memory Configuration objects one at a time
        by loading the atomic numbers, positions, cells, and PBCs.

        Args:

            ids (list or 'all'):
                A list of string IDs specifying which Configurations to return.
                If 'all', returns all of the configurations in the database.

            generator (bool):
                If true, this function becomes a generator which only returns
                the configurations one at a time. This is useful if the
                configurations can't all fit in memory at the same time. Default
                is False.

            verbose (bool):
                If True, prints progress bar

        Returns:

            configurations (iterable):
                A list or generator of the re-constructed configurations
        """

        if generator:
            return self._get_configurations_gen(ids, verbose=verbose)
        else:
            return self._get_configurations(ids, verbose=verbose)


    def _get_configurations(self, ids, verbose=False):
        if ids == 'all':
            # ids = [
            #     ds.asstr()[0]
            #     for ds in self.get_data('configurations/ids').values()
            # ]

            ids = list(self['configurations/ids/data'].keys())

        configurations = []

        for co_id in tqdm(
            ids,
            desc='Loading configurations',
            disable=not verbose
            ):
            # atoms = Atoms(
            #     symbols=self.get_data(
            #         'configurations/atomic_numbers'
            #     )[self[f'configurations/atomic_numbers/slices/{co_id}'][()]],

            #     positions=self.get_data(
            #         'configurations/positions'
            #     )[self[f'configurations/positions/slices/{co_id}'][()]],

            #     cell=self.get_data(
            #         'configurations/cells'
            #     )[self[f'configurations/cells/slices/{co_id}'][()]],

            #     pbc=self.get_data(
            #         'configurations/pbcs'
            #     )[self[f'configurations/pbcs/slices/{co_id}'][()]],
            # )

            # atoms.info[ATOMS_NAME_FIELD] = set(self.get_data(
            #     'configurations/names'
            # )[self[f'configurations/names/slices/{co_id}'][()]].asstr())

            # atoms.info[ATOMS_LABELS_FIELD] = set(self.get_data(
            #     'configurations/labels'
            # )[self[f'configurations/labels/slices/{co_id}'][()]].asstr())

            # atoms.info[ATOMS_LAST_MODIFIED_FIELD] = self.get_data(
            #     'configurations/last_modified'
            # )[self[f'configurations/last_modified/slices/{co_id}'][()]].asstr()[()]

            atoms = Atoms(
                symbols=self[f'configurations/atomic_numbers/data/{co_id}'],
                positions=self[f'configurations/positions/data/{co_id}'],
                cell=self[f'configurations/cells/data/{co_id}'],
                pbc=self[f'configurations/pbcs/data/{co_id}'],
            )

            atoms.info[ATOMS_NAME_FIELD] = set(self[f'configurations/names/data/{co_id}'].asstr()[()])
            atoms.info[ATOMS_LABELS_FIELD] = set(self[f'configurations/labels/data/{co_id}'].asstr()[()])
            atoms.info[ATOMS_LAST_MODIFIED_FIELD] = self[f'configurations/last_modified/data/{co_id}'].asstr()[()]

            configurations.append(Configuration.from_ase(atoms))

        return configurations



    def _get_configurations_gen(self, ids, verbose=False):
        if ids == 'all':
            ids = [
                ds.asstr()[0]
                for ds in self.get_data('configurations/ids').values()
            ]

        for co_id in tqdm(
            ids,
            desc='Loading configurations',
            disable=not verbose
            ):
            # atoms = Atoms(
            #     symbols=self.get_data(
            #         'configurations/atomic_numbers'
            #     )[self[f'configurations/atomic_numbers/slices/{co_id}'][()]],

            #     positions=self.get_data(
            #         'configurations/positions'
            #     )[self[f'configurations/positions/slices/{co_id}'][()]],

            #     cell=self.get_data(
            #         'configurations/cells'
            #     )[self[f'configurations/cells/slices/{co_id}'][()]],

            #     pbc=self.get_data(
            #         'configurations/pbcs'
            #     )[self[f'configurations/pbcs/slices/{co_id}'][()]],
            # )

            # atoms.info[ATOMS_NAME_FIELD] = set(self.get_data(
            #     'configurations/names'
            # )[self[f'configurations/names/slices/{co_id}'][()]].asstr())

            # atoms.info[ATOMS_LABELS_FIELD] = set(self.get_data(
            #     'configurations/labels'
            # )[self[f'configurations/labels/slices/{co_id}'][()]].asstr())

            # atoms.info[ATOMS_LAST_MODIFIED_FIELD] = self.get_data(
            #     'configurations/last_modified'
            # )[self[f'configurations/last_modified/slices/{co_id}'][()]].asstr()[()]

            atoms = Atoms(
                symbols=self[f'configurations/atomic_numbers/data/{co_id}'][()],
                positions=self[f'configurations/positions/data/{co_id}'][()],
                cell=self[f'configurations/cells/data/{co_id}'][()],
                pbc=self[f'configurations/pbcs/data/{co_id}'][()],
            )

            atoms.info[ATOMS_NAME_FIELD] = set(self[f'configurations/names/data/{co_id}'].asstr()[()])
            atoms.info[ATOMS_LABELS_FIELD] = set(self[f'configurations/labels/data/{co_id}'].asstr()[()])
            atoms.info[ATOMS_LAST_MODIFIED_FIELD] = self[f'configurations/last_modified/data/{co_id}'].asstr()[()]

            yield Configuration.from_ase(atoms)


    def concatenate_configurations(self):
        """
        Concatenates the atomic_numbers, positions, cells, and pbcs groups in
        /configurations.
        """
        self.concatenate_group('configurations/atomic_numbers')
        self.concatenate_group('configurations/positions')
        self.concatenate_group('configurations/cells')
        self.concatenate_group('configurations/pbcs')


    def insert_property_definition(self, definition):
        """
        Inserts a new property definition into the database. Checks that
        definition is valid, then builds all necessary groups in
        :code:`/root/properties`.

        Args:

            definition (dict):
                The map defining the property. See the example below, or the
                `OpenKIM Properties Framework <https://openkim.org/doc/schema/properties-framework/>`_
                for more details.

        Example definition:

        ..code-block:: python

            property_definition = {
                'property-id': 'default',
                'property-title': 'A default property used for testing',
                'property-description': 'A description of the property',
                'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                'nd-diff-shape': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                'nd-diff-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
            }
        """

        dummy_dict = deepcopy(definition)

        # Spoof if necessary
        if VALID_KIM_ID.match(dummy_dict['property-id']) is None:
            # Invalid ID. Try spoofing it
            dummy_dict['property-id'] = 'tag:@,0000-00-00:property/'
            dummy_dict['property-id'] += definition['property-id']
            warnings.warn(f"Invalid KIM property-id; Temporarily renaming to {dummy_dict['property-id']}")

        check_property_definition(dummy_dict)

        group = self['properties'].require_group(definition['property-id'])
        group.attrs['definition'] = json.dumps(definition)

        now = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        group.attrs['last_modified'] = now

        ignore_fields = {
            'property-id', 'property-title', 'property-description'
        }

        for subgroup in ['configuration_ids', 'settings_ids']:
            g = group.create_group(subgroup)
            g.create_group('data', track_order=True)
            g.create_group('slices')
            g.attrs['concatenated'] = False

        for field in set(definition.keys()) - ignore_fields:
            g = group.create_group(field)
            g.create_group('data', track_order=True)
            g.create_group('slices')
            g.attrs['concatenated'] = False


    def get_property_definition(self, name):
        """
        Returns:
            A dictionary with two keys:
                'last_modified': a datetime string
                'definition': the dictionary form of the property definition
        """
        return {
            'last_modified': self[f'properties/{name}'].attrs['last_modified'],
            'definition': json.loads(
                self[f'properties/{name}'].attrs['definition']
            )
        }


    def insert_property_settings(self, pso_object):
        """
        Inserts a new property settings object into the database by creating
        and populating the necessary groups in :code:`/root/property_settings`.

        Args:

            pso_object (PropertySettings)
                The :class:`~colabfit.tools.property_settings.PropertySettings`
                object to insert into the database.


        Returns:

            pso_id (str):
                The ID of the inserted property settings object. Equals the hash
                of the object.
        """
        pso_id = str(hash(pso_object))

        if pso_id in self['property_settings']:
            return pso_id

        g = self['property_settings'].create_group(pso_id)

        now = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        g.attrs['last_modified']    = now
        g.attrs['method']           = pso_object.method
        g.attrs['description']      = pso_object.description
        g.attrs['labels']           = json.dumps(list(pso_object.labels))

        g = g.create_group('files')
        for (fname, contents) in pso_object.files:
            g[fname] = contents

        return pso_id


    def get_property_settings(self, pso_id):
        """
        Returns:
            A dictionary with two keys:
                'last_modified': a datetime string
                'settings': the PropertySettings object with the given ID
        """
        g = self[f'property_settings/{pso_id}']

        return {
            'last_modified': g.attrs['last_modified'],
            'settings': PropertySettings(
                method=g.attrs['method'],
                description=g.attrs['description'],
                labels=set(json.loads(g.attrs['labels'])),
                files=[(fname, ds.asstr()[()]) for fname, ds in g['files'].items()]
            )
        }


    def get_properties(self, ids):
        """
        This function should take a list of PR IDs, and re-construct the
        properties from them.
        """
        pass


    def insert_configuration_set(self, ids, description=''):
        """
        Inserts the configuration set of IDs to the database.

        Args:

            ids (list or str):
                The IDs of the configurations to include in the configuartion
                set.

            description (str, optional):
                A human-readable description of the configuration set.
        """

        if isinstance(ids, str):
            ids = [ids]

        cs_hash = sha512()
        for i in sorted(ids):
            cs_hash.update(str(i).encode('utf-8'))

        cs_id = str(int(cs_hash.hexdigest()[:16], 16)-HASH_SHIFT)

        # Check for duplicates
        if cs_id in self['configuration_sets']:
            return cs_id

        # Make sure all of the configurations exist
        for co_id in ids:
            if co_id not in self['configurations/ids/data']:
                raise MissingEntryError(
                    "The configuration with ID '{}' is not in the "\
                    "database".format(co_id)
                )

        g = self['configuration_sets'].create_group(cs_id)
        g.attrs['description'] = description
        # g = g.create_group('ids')
        # g.attrs['concatenated'] = True
        # g = g.create_group('data', track_order=True)
        # g.create_dataset(
        #     name='_root_concatenated',
        #     data=np.array(ids, dtype=STRING_DTYPE_SPECIFIER)
        # )

        g.attrs['configuration_ids'] = ids

        now = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        g.attrs['last_modified'] = now

        return cs_id


    def insert_dataset(
        self, cs_ids, pr_ids,
        authors, links, description,
        ):
        """
        Inserts a dataset into the database.

        Args:

            cs_ids (list or str):
                The IDs of the configuration sets to link to the dataset.

            pr_ids (list or str):
                The IDs of the properties to link to the dataset

            authors (list or str or None):
                The names of the authors of the dataset. If None, then no
                authors are added.

            links (list or str or None):
                External links (e.g., journal articles, Git repositories, ...)
                to be associated with the dataset. If None, then no links are
                added.

            description (str or None):
                A human-readable description of the dataset. If None, then not
                description is added.
        """

        if isinstance(cs_ids, str):
            cs_ids = [cs_ids]

        if isinstance(pr_ids, str):
            pr_ids = [pr_ids]

        if isinstance(authors, str):
            authors = [authors]

        if isinstance(links, str):
            links = [links]

        ds_hash = sha512()
        for ci in sorted(cs_ids):
            ds_hash.update(str(ci).encode('utf-8'))
        for pi in sorted(pr_ids):
            ds_hash.update(str(pi).encode('utf-8'))

        ds_id = str(int(ds_hash.hexdigest()[:16], 16)-HASH_SHIFT)

        # Check for duplicates
        if ds_id in self['datasets']:
            return ds_id

        # Make sure all of the configuration sets and properties exist
        for cs_id in cs_ids:
            if cs_id not in self['configuration_sets']:
                raise MissingEntryError(
                    "The configuration set with ID '{}' is not in the "\
                    "database".format(cs_id)
                )

        for pr_id in pr_ids:
            if pr_id not in self['properties/ids/data']:
                raise MissingEntryError(
                    "The configuration set with ID '{}' is not in the "\
                    "database".format(pr_id)
                )

        g = self['datasets'].create_group(ds_id)
        g.attrs['authors'] = authors
        g.attrs['links'] = links
        g.attrs['description'] = description

        g.attrs['configuration_set_ids'] = cs_ids
        # g = g.create_group('configuration_set_ids')
        # g.attrs['concatenated'] = True
        # g = g.create_group('data', track_order=True)
        # g.create_dataset(
        #     name='_root_concatenated',
        #     data=np.array(cs_ids, dtype=STRING_DTYPE_SPECIFIER)
        # )

        g.attrs['property_ids'] = pr_ids
        # g = g.create_group('property_ids')
        # g.attrs['concatenated'] = True
        # g = g.create_group('data', track_order=True)
        # g.create_dataset(
        #     name='_root_concatenated',
        #     data=np.array(pr_ids, dtype=STRING_DTYPE_SPECIFIER)
        # )

        now = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        self[f'datasets/{ds_id}'].attrs['last_modified']    = now

        return ds_id

class ConcatenationException(Exception):
    pass

class InvalidGroupError(Exception):
    pass

class MissingEntryError(Exception):
    pass


__version__ = "0.0.2"
__author__ = "ColabFit"
__credit__ = "University of Minnesota"

HASH_LENGTH = 12
HASH_SHIFT = 0
# HASH_SHIFT = 2**63

ID_FORMAT_STRING = "{}_{}_{:0d}"

MAX_STRING_LENGTH = 255
STRING_DTYPE_SPECIFIER = f"S{MAX_STRING_LENGTH}"

# _DBASE_COLLECTION           = '_databases'
_DATABASE_NAME = "colabfit_database"
_CONFIGS_COLLECTION = "configurations"
_PROPOBJECT_COLLECTION = "property_objects"
_PROPDEFS_COLLECTION = "property_definitions"
_DATAOBJECT_COLLECTION = "data_objects"
_CONFIGSETS_COLLECTION = "configuration_sets"
_DATASETS_COLLECTION = "datasets"

SHORT_ID_STRING_NAME = "id"
EXTENDED_ID_STRING_NAME = "extended_id"

ATOMS_NAME_FIELD = "_name"
ATOMS_LABELS_FIELD = "_labels"
ATOMS_LAST_MODIFIED_FIELD = "_last_modified"
ATOMS_CONSTRAINTS_FIELD = "_constraints"

DEFAULT_PROPERTY_NAME = (
    "configuration-nonorthogonal-periodic-3d-cell-fixed-particles-fixed"
)

from ase.units import create_units

UNITS = create_units("2014")

# # Make GPa the base unit
# UNITS["bar"] = 1e-4  # bar to GPa
# UNITS["kilobar"] = 1e-1  # kilobar to GPa
# UNITS["pascal"] = 1e-9  # pascal to GPa
# UNITS["GPa"] = 1

# UNITS["angstrom"] = UNITS["Ang"]

# OPENKIM_PROPERTY_UNITS = {
#     "energy": "eV",
#     "forces": "eV/angstrom",
#     "stress": "GPa",
#     "unrelaxed-potential-energy": "eV",
#     "unrelaxed-potential-forces": "eV/angstrom",
#     "unrelaxed-cauchy-stress": "GPa",
# }

EDN_KEY_MAP = {
    "energy": "unrelaxed-potential-energy",
    "forces": "unrelaxed-potential-forces",
    "stress": "unrelaxed-cauchy-stress",
    "virial": "unrelaxed-cauchy-stress",
}
